# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List
import unittest
import sys, os

if __name__ == "__main__":
    sys.path.append("..")
    sys.path.append(".")
import numpy as np
from fractions import Fraction

from pyzx_param.routing import architecture
from pyzx_param.routing.architecture import create_architecture
from pyzx_param.routing.cnot_mapper import ElimMode
from pyzx_param.routing.parity_maps import CNOT_tracker, Parity
from pyzx_param.linalg import Mat2
from pyzx_param.routing.phase_poly import PhasePoly
from pyzx_param.circuit import Circuit, HAD, CNOT, ZPhase, XPhase
from pyzx_param.tensor import compare_tensors
from pyzx_param.extract import permutation_as_swaps

SEED = 42


class TestPhasePoly(unittest.TestCase):
    def setUp(self):
        np.random.seed(SEED)

        self.n_tests = 1
        self.n_qubits = 5
        name = architecture.LINE
        self.architecture = create_architecture(name, n_qubits=self.n_qubits)
        # Define some circuits to work with
        folder = "circuits/steiner/" + str(self.n_qubits) + "qubits/"
        try:
            cnot_options = next(os.walk(folder))
        except StopIteration:
            # There are no circuits to test with
            return
        n_cnots = cnot_options[1]
        self.circuit = []
        n_phase_layers = 5
        self.n_phase_layers = n_phase_layers

        def filename():
            return os.path.join(
                *[
                    folder,
                    n_cnots[np.random.choice(len(n_cnots))],
                    "Original" + str(np.random.choice(20)) + ".qasm",
                ]
            )

        for _ in range(self.n_tests):
            c = Circuit.from_qasm_file(filename())
            for _ in range(n_phase_layers):
                for i in range(self.n_qubits):
                    if (
                        np.random.choice(2, p=[0.2, 0.8]) == 1
                    ):  # Pick H gate with chance .2
                        phase = np.random.choice([1, -1]) * Fraction(
                            1, int(np.random.choice([1, 2, 4]))
                        )
                        c.add_gate(ZPhase(target=i, phase=phase))
                    else:
                        c.add_gate(HAD(target=i))
                c.add_circuit(Circuit.from_qasm_file(filename()))
            self.circuit.append(c)

    def assertGates(self, circuit):
        for gate in circuit.gates:
            if hasattr(gate, "name") and gate.name == "CNOT":
                edge = (gate.control, gate.target)
                self.assertTrue(
                    edge in self.architecture.graph.edges()
                    or tuple(reversed(edge)) in self.architecture.graph.edges()
                )

    def assertMat2Equal(self, m1, m2, triangle=False):
        if triangle:
            self.assertListEqual(
                *[
                    [
                        m.data[i, j]
                        for i in range(self.n_qubits)
                        for j in range(self.n_qubits)
                        if i >= j
                    ]
                    for m in [m1, m2]
                ]
            )
        else:
            self.assertNdArrEqual(m1.data, m2.data)

    def assertNdArrEqual(self, a1, a2):
        if isinstance(a1, list):
            if isinstance(a2, list):
                self.assertListEqual(a1, a2)
            else:
                self.assertListEqual(a1, a2.tolist())
        else:
            if isinstance(a2, list):
                self.assertListEqual(a1.tolist(), a2)
            else:
                self.assertListEqual(a1.tolist(), a2.tolist())

    def assertCircuitEquivalent(self, c1, c2):
        the_same = compare_tensors(c1, c2)
        self.assertTrue(the_same)

    def assertPhasePolyEqual(self, p1, p2):
        # self.assertDictEqual(p1.xphases, p2.xphases)
        self.assertDictEqual(p1.zphases, p2.zphases)
        self.assertListEqual(p1.out_par, p2.out_par)

    def assertPartitionEqual(self, sets, partitions: List[List[Parity]]):
        flat_partition = [p for subset in partitions for p in subset]
        # Are all sets partitioned?
        for s in sets:
            self.assertIn(s, flat_partition)
        extra_parities = []
        for p in flat_partition:
            if p not in sets:
                extra_parities.append(p)
        for p in extra_parities:
            self.assertTrue(p.count() == 1)
        # self.assertCountEqual(extra_parities, list(set(extra_parities))) # No duplicates
        # [self.assertIn(p, sets) for p in flat_partition]
        # Is every partition a set of independent parities?
        for pt in partitions:
            self.assertTrue(PhasePoly._independent(pt))

    def assertFinalParityEqual(self, c1, c2):
        old_cnots = CNOT_tracker(c1.qubits)
        new_cnots = CNOT_tracker(c1.qubits)
        for gate in c1.gates:
            if isinstance(gate, CNOT):
                old_cnots.row_add(gate.control, gate.target)
        for gate in c2.gates:
            if isinstance(gate, CNOT):
                new_cnots.row_add(gate.control, gate.target)
        self.assertCircuitEquivalent(old_cnots, new_cnots)
        return old_cnots, new_cnots

    """ Matroid partitioning is not a good algorithm
    def test_partitions(self):
        for i in range(self.n_tests):
            with self.subTest(i=i):
                circuit = Circuit(self.circuit[i].qubits)
                for gate in self.circuit[i].gates:
                    if isinstance(gate, HAD):
                        break
                    circuit.add_gate(gate)
                # Check if the phase poly is created properly
                phase_poly = PhasePoly.fromCircuit(circuit)
                # Check the partitions
                partitions = phase_poly.partition(skip_output_parities=False)
                self.assertPartitionEqual(phase_poly.all_parities, partitions)
    """

    def test_phase_poly_parity_creation(self):
        for i in range(self.n_tests):
            with self.subTest(i=i):
                n_qubits = self.circuit[i].qubits
                circuit = Circuit(n_qubits)
                for gate in self.circuit[i].gates:
                    if isinstance(gate, HAD):
                        break
                    circuit.add_gate(gate)
                # Check if the phase poly is created properly
                phase_poly = PhasePoly.fromCircuit(circuit)
                old_cnots = CNOT_tracker(n_qubits)
                for gate in circuit.gates:
                    if isinstance(gate, CNOT):
                        old_cnots.row_add(gate.control, gate.target)
                original_out_par = [
                    Parity(row) for row in old_cnots.matrix.data
                ]
                # Check if the output parities are the same
                self.assertListEqual(original_out_par, phase_poly.out_par)

    def test_phase_poly_phase_creation(self):
        for i in range(self.n_tests):
            with self.subTest(i=i):
                n_qubits = self.circuit[i].qubits
                circuit = Circuit(n_qubits)
                had_phase = False
                for gate in self.circuit[i].gates:
                    if isinstance(gate, HAD):
                        break
                    elif isinstance(gate, ZPhase) or isinstance(gate, XPhase):
                        had_phase = True
                    elif had_phase and isinstance(gate, CNOT):
                        break
                    circuit.add_gate(gate)
                # Check if the phase poly is created properly
                phase_poly = PhasePoly.fromCircuit(circuit)
                old_cnots = CNOT_tracker(n_qubits)
                for gate in circuit.gates:
                    if isinstance(gate, CNOT):
                        old_cnots.row_add(gate.control, gate.target)
                original_out_par = [
                    "".join([str(v) for v in row]) for row in old_cnots.matrix.data
                ]
                # Check if the output parities are the same
                self.assertListEqual(original_out_par, phase_poly.out_par)
                # Check if the original phase poly is the same
                phase_poly_parities = list(
                    phase_poly.zphases.keys()
                )  # + list(phase_poly.xphases.keys())
                for parity in phase_poly_parities:
                    # Check  if all phase poly parities are in the original
                    self.assertIn(parity, original_out_par)

    def test_phase_poly_routed(self):
        for i in range(self.n_tests):
            with self.subTest(i=i):
                c = Circuit(self.circuit[i].qubits)
                for gate in self.circuit[i].gates:
                    if isinstance(gate, HAD):
                        break
                    c.add_gate(gate)
                self.do_phase_poly(
                    c.copy(), ElimMode.STEINER_MODE, self.architecture, routed=True
                )

    def do_phase_poly(
        self, circuit, mode, architecture, routed=False, tensor_compare=False
    ):
        phase_poly = PhasePoly.fromCircuit(circuit)
        # TODO: test the other synthesis functions
        # for func in [phase_poly.matroid_synth, phase_poly.gray_synth, phase_poly.rec_gray_synth, phase_poly.Ariannes_synth]:
        for func in [phase_poly.Ariannes_synth]:
            with self.subTest(i=func):
                # Check the synthesized circuit
                new_circuit, initial_perm, output_perm = func(
                    mode=mode, architecture=architecture, full_reduce=True
                )
                if routed:
                    self.assertGates(new_circuit)
                adjusted_circuit = self.apply_perms(
                    new_circuit, initial_perm, output_perm
                )
                new_phase_poly = PhasePoly.fromCircuit(adjusted_circuit)
                # Check if the phasepolys are the same
                self.assertPhasePolyEqual(phase_poly, new_phase_poly)
                self.assertFinalParityEqual(circuit, adjusted_circuit)
                # Check if the circuits are the same
                self.assertCircuitEquivalent(adjusted_circuit, circuit)

    def test_tensor_compare(self):
        n_qubits = 2
        old_qubits = self.n_qubits
        self.n_qubits = n_qubits
        circuit = Circuit(n_qubits)
        n_cnots = 1
        for _ in range(n_cnots):
            circuit.add_gate("CNOT", *np.random.choice(n_qubits, 2, False))
        for i in range(self.n_qubits):
            phase = np.random.choice([1, -1]) * Fraction(
                1, int(np.random.choice([1, 2, 4]))
            )
            circuit.add_gate(ZPhase(target=i, phase=phase))
        for _ in range(n_cnots):
            circuit.add_gate("CNOT", *np.random.choice(n_qubits, 2, False))
        # for i in range(self.n_qubits):
        #    phase = np.random.choice([1,-1])*Fraction(1, int(np.random.choice([1,2,4])))
        #    circuit.add_gate(ZPhase(target=i, phase=phase))
        # for _ in range(n_cnots):
        #    circuit.add_gate("CNOT", *np.random.choice(n_qubits, 2, False))
        self.do_phase_poly(
            circuit.copy(),
            ElimMode.STEINER_MODE,
            None,
            routed=False,
            tensor_compare=True,
        )
        self.n_qubits = old_qubits

    def apply_perms(self, circuit, initial_perm, output_perm):
        adjusted_circuit = Circuit(self.n_qubits)
        # Undo the initial permutation
        for q1, q2 in permutation_as_swaps({v: k for k, v in enumerate(initial_perm)}):
            adjusted_circuit.add_gate(CNOT(q1, q2))
            adjusted_circuit.add_gate(CNOT(q2, q1))
            adjusted_circuit.add_gate(CNOT(q1, q2))
        # Do the circuit
        for gate in circuit.gates:
            adjusted_circuit.add_gate(gate)
        # Realise the output permutation
        for q1, q2 in permutation_as_swaps({k: v for k, v in enumerate(output_perm)}):
            adjusted_circuit.add_gate(CNOT(q1, q2))
            adjusted_circuit.add_gate(CNOT(q2, q1))
            adjusted_circuit.add_gate(CNOT(q1, q2))
        return adjusted_circuit


if __name__ == "__main__":
    print("Testing PhasePoly")
    unittest.main()



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from multiprocessing import cpu_count
from multiprocessing.pool import Pool
from typing import Any, List, Optional, Tuple
import numpy as np


class GeneticAlgorithm:
    """
    A genetic algorithm for optimising permutations based on a fitness function.
    """
    def __init__(
        self,
        population_size: int,
        crossover_prob: float,
        mutation_prob: float,
        fitness_func,
        maximize: bool = False,
    ):
        """
        Creates and returns a genetic algorithm.

        :param population_size: Number of individuals in the population
        :param crossover_prob: Probability of crossover between individuals
        :param mutation_prob: Probability of mutation for an offspring
        :param fitness_func: Function to evaluate fitness of permutations
        :param maximize: True, Maximise the fitness, False, Minimise the Fitness, default False 
        """
        self.population_size = population_size
        self.negative_population_size = int(np.sqrt(population_size))
        self.crossover_prob = crossover_prob
        self.mutation_prob = mutation_prob
        self.fitness_func = fitness_func
        self._sort = lambda l: l.sort(key=lambda x: x[1], reverse=maximize)
        self.maximize = maximize
        self.n_qubits = 0
        self.population: List[Tuple[List[int], Any]] = []

    def _select(self):
        """
        Selects two parent indices from the population based on fitness-proportional selection.
        
        :return: A random two parent indices from the population
        """
        fitness_scores = [f for c, f in self.population]
        total_fitness = sum(fitness_scores)
        if self.maximize:
            selection_chance = [f / total_fitness for f in fitness_scores]
        else:
            max_fitness = max(fitness_scores) + 1
            adjusted_scores = [max_fitness - f for f in fitness_scores]
            adjusted_total = sum(adjusted_scores)
            selection_chance = [f / adjusted_total for f in adjusted_scores]
        return np.random.choice(
            self.population_size, size=2, replace=False, p=selection_chance
        )

    def _create_population(self, n):
        """
        Initialises the population with random permutations with size n and evaluates their fitness.
        Also creates a list of the weakest individuals - negative population.

        :param n: The size of the random permutation
        """
        population = [np.random.permutation(n) for _ in range(self.population_size)]
        self.population = [
            (list(chromosome), self.fitness_func(chromosome))
            for chromosome in population
        ]
        self._sort(self.population)
        self.negative_population = self.population[-self.negative_population_size :]

    def find_optimum(
        self, n_qubits, n_generations, initial_order=None, n_child=None, continued=False
    ):
        """
        Runs the genetic algorithm to find the best permutation over a number of generations.

        :param n_qubits: The number of qubits (number of elements in each permutation)
        :param n_generations: The number of generations
        :param initial_order: Initial permuation to start from, if None, creates population with the number of qubits, default None
        :param n_child: Number of children to generate per generation, default None
        :param continued: True, continue to previous population, default False
        :return: The best permutation
        """
        self.n_qubits = n_qubits
        partial_solution = False
        if not continued or not self.population:
            if initial_order is None:
                self._create_population(n_qubits)
            elif n_qubits < len(initial_order):
                self._create_population(initial_order[:n_qubits])
                partial_solution = True
            else:
                self._create_population(initial_order)

        if n_child is None:
            n_child = self.population_size

        for _ in range(n_generations):
            self._update_population(n_child)
        if partial_solution and initial_order is not None:
            return self.population[0] + initial_order[n_qubits:]
        return self.population[0][0]

    def _add_children(self, children):
        """
        Adds new children to the population and updates fitness scores.

        :param children: The children to be added to the population
        """
        n_child = len(children)
        self.population.extend(
            [(child, self.fitness_func(child)) for child in children]
        )
        self._sort(self.population)
        self.negative_population.extend(self.population[-n_child:])
        self.negative_population = [
            self.negative_population[i]
            for i in np.random.choice(
                self.negative_population_size + n_child,
                size=self.negative_population_size,
                replace=False,
            )
        ]
        self.population = self.population[: self.population_size]

    def _update_population(self, n_child: int):
        """
        Generates a new generation of children and integrates them into the population.

        :param n_child: The number of children
        """
        children = []
        # Create a child from weak parents to avoid local optima
        p1, p2 = np.random.choice(self.negative_population_size, size=2, replace=False)
        child = self._crossover(
            self.negative_population[p1][0], self.negative_population[p2][0]
        )
        children.append(child)
        for _ in range(n_child):
            if np.random.random() < self.crossover_prob:
                p1, p2 = self._select()
                child = self._crossover(self.population[p1][0], self.population[p2][0])
                if np.random.random() < self.mutation_prob:
                    child = self._mutate(child)
                children.append(child)
        self._add_children(children)

    def _crossover(self, parent1, parent2):
        """
        Performs ordered crossover between two parents.

        :param parent1: The first parent
        :param parent2: The second parent
        :return: The child of the 2 parents
        """
        crossover_start = np.random.choice(int(self.n_qubits / 2))
        crossover_length = np.random.choice(self.n_qubits - crossover_start)
        crossover_end = crossover_start + crossover_length
        child = -1 * np.ones_like(parent1)
        child[crossover_start:crossover_end] = parent1[crossover_start:crossover_end]
        child_idx = 0
        for parent_gen in parent2:
            if child_idx == crossover_start:  # skip over the parent1 part in child
                child_idx = crossover_end
            if parent_gen not in child:  # only add new genes
                child[child_idx] = parent_gen
                child_idx += 1
        return child

    def _mutate(self, parent):
        """
        Applies mutation by swapping two random elements in the permutation.

        :param parent: The parent to mutate
        :return: A mutated parent
        """
        gen1, gen2 = np.random.choice(len(parent), size=2, replace=False)
        _ = parent[gen1]
        parent[gen1] = parent[gen2]
        parent[gen2] = _
        return parent

class ParticleSwarmOptimization:
    """
    Optimizer based on the particle swarm optimization algorithm.
    """

    def __init__(
        self,
        swarm_size: int,
        step_func, # type 'StepFunction'
        s_best_crossover: float,
        p_best_crossover: float,
        mutation: float,
        maximize: bool = False,
        n_threads: Optional[int] = None,
    ):
        """
        Setup the optimizer

        :param swarm_size: Swarm size for the swarm optimization.
        :param step_function: The to progress the swarm.
        :param s_best_crossover: The crossover percentage with the best particle in the swarm. Must be between 0.0 and 1.0.
        :param p_best_crossover: The crossover percentage with the personal best of a particle. Must be between 0.0 and 1.0.
        :param mutation: The mutation percentage of a particle. Must be between 0.0 and 1.0.
        :param maximize: Whether to maximize the fitness function.
        :param n_threads: Number of threads to use for the optimization. If None, use all available threads.
        """
        self.step_func = step_func
        self.size = swarm_size
        self.s_crossover = s_best_crossover
        self.p_crossover = p_best_crossover
        self.mutation = mutation
        self.best_particle = None
        self.maximize = maximize
        self.swarm: List[Particle] = []
        self.pool: Optional[Pool] = None
        n_threads = (
            min(n_threads, cpu_count()) if n_threads is not None else cpu_count()
        )
        if n_threads > 1:
            self.pool = Pool(n_threads)

    def __getstate__(self):
        """
        Prepares the object state for pickling by removing non-serialisable fields.

        :return: The state dictionary
        """
        state = self.__dict__.copy()
        # Don't pickle baz
        # del state["fitness_func"]
        del state["pool"]
        return state

    def __setstate__(self, state):
        """
        Restores the object state after unpickling

        :param state: The state to be restored to the dictionary
        """
        self.__dict__.update(state)
        # Add baz back since it doesn't exist in the pickle
        # self.fitness_func = None
        self.pool = None

    def _create_swarm(self, n):
        """
        Initializes the swarm with random permutations.

        :param n: The number of qubits in a particle
        """

        self.swarm = [
            Particle(
                n,
                self.step_func,
                self.s_crossover,
                self.p_crossover,
                self.mutation,
                self.maximize,
                id=i,
            )
            for i in range(self.size)
        ]
        # Start with 1 particle with initial permutation
        self.swarm[0].current = np.arange(n)

    def find_optimum(self, n_qubits, n_steps, quiet=True, close_pool=True):
        """
        Creates a swarm of n-qubits and determines the optimum fitness solution for a given number of steps

        :param n_qubits: The number of qubits
        :param n_steps: The number of steps
        :param quiet: Whether to show updates on the iteration and fitness score as it iterates, default True
        :param close_pool: Whether to close and join the pool after finding the optimum, default True
        :return: The optimum solution for swarm
        """
        self._create_swarm(n_qubits)
        self.best_particle = self.swarm[0]
        for i in range(n_steps):
            self._update_swarm()
            if not quiet:
                print(
                    "PSO - Iteration",
                    i,
                    "best fitness:",
                    self.best_particle.best,
                    self.best_particle.best_point,
                )
        if close_pool and self.pool:
            self.pool.close()
            self.pool.join()
        return self.best_particle.best_solution

    @staticmethod
    def particle_update_func(args):
        """
        Update a single particle in the swarm using the current global best particle.

        :param args: Tuple of swarm_best and particle, where swarm_best is the current best performing particle and the particle is the particle to be updated
        :return: The next evolution of the particle
        """
        swarm_best, p = args
        p.step(swarm_best)
        return p

    def _update_swarm(self):
        """
        Update the state of all particles in the swarm, after updating the method finds the best particle in the swarm.
        """
        if self.pool is not None:
            self.swarm = self.pool.map(
                self.particle_update_func, [(self.best_particle, p) for p in self.swarm]
            )
        else:
            self.swarm = [
                self.particle_update_func((self.best_particle, p)) for p in self.swarm
            ]
        if self.maximize:
            top = max(
                self.swarm, key=lambda p: p.best if p.best is not None else -np.inf
            )
        else:
            top = min(
                self.swarm, key=lambda p: p.best if p.best is not None else np.inf
            )
        if self.best_particle is None or top.compare(self.best_particle.best):
            self.best_particle = top


class Particle:
    """
    Represents a single particle in a swarm.
    """
    def __init__(
        self,
        size,
        step_func,
        s_best_crossover,
        p_best_crossover,
        mutation,
        maximize=False,
        id=None,
    ):
        """
        Creates and returns a single particle.

        :param size: The number of qubits in the permutation space
        :param step_func: A callable that evaluates the current position and returns (new_position, solution, fitness)
        :param s_best_crossover: Proportion of genes inherited from the swarms best particle
        :param p_best_crossover: Proportion of genes inherited from the particles own best known position
        :param mutation: Mutation rate as a proportion of the permatation length
        :param maximize: True, Maximise the fitness, False, Minimise the fitness, default False 
        :param id: Id for the particle, default None
        """
        self.step_func = step_func
        self.size = size
        self.current = np.random.permutation(size)
        self.best_point = self.current
        self.best = None
        self.best_solution = None
        self.s_crossover = int(s_best_crossover * size)
        self.p_crossover = int(p_best_crossover * size)
        self.mutation = int(mutation * size)
        self.maximize = maximize
        self.id = id

    def compare(self, x):
        """
        Compare a new fitness score to the current best.

        :param x: New score to compare
        :return: True, if it is the better than the best score
        """
        if x is None:
            return True
        if self.maximize:
            return x < self.best
        else:
            return x > self.best

    def step(self, swarm_best):
        """
        Preform one optimisation step for the particle.

        :param swarm_best: The best particle in the swarm
        :return: True, a better solution was found, False, no better solution was found
        """
        new, solution, fitness = self.step_func(self.current)
        is_better = self.best is None or not self.compare(fitness)
        if is_better:
            self.best = fitness
            self.best_point = self.current
            self.best_solution = solution
        elif all([self.current[i] == n for i, n in enumerate(new)]):
            new = self._mutate(self.current)
            new = self._crossover(new, self.best_point, self.p_crossover)
            new = self._crossover(new, swarm_best.best_point, self.s_crossover)
            # Sanity check TODO can be removed!
            if any([i not in new for i in range(self.size)]):
                raise Exception(
                    "The new particle point is not a permutation anymore!"
                    + str(self.current)
                )
        self.current = new
        return is_better

    def _mutate(self, particle):
        """
        Randomly mutate the particle by permuting a subset of its genes.

        :param particle: The current particle permutation
        :return: A mutated particle permutation
        """
        new_particle = particle.copy()
        m_idxs = np.random.choice(self.size, size=self.mutation, replace=False)
        m_perm = np.random.permutation(self.mutation)
        for old_i, new_i in enumerate(m_perm):
            new_particle[m_idxs[old_i]] = particle[m_idxs[new_i]]
        return new_particle

    def _crossover(self, particle, best_particle, n):
        """
        Preform a crossover between this particle and the best current particle.

        :param particle: The current particle permutation
        :param best_particle: The highest scoring particle permutation
        :param n: The number of genes to crossover
        :return: A new mutated particle permutation resulting from the crossover
        """
        cross_idxs = np.random.choice(self.size, size=n, replace=False)
        new_particle = -1 * np.ones_like(particle)
        new_particle[cross_idxs] = best_particle[cross_idxs]
        idx = 0
        for i, gen in enumerate(new_particle):
            if gen == -1:  # skip over the parent1 part in child
                while particle[idx] in new_particle:
                    idx += 1
                    if idx == len(particle):
                        break
                if idx < len(particle):
                    new_particle[i] = particle[idx]
        return new_particle

if __name__ == "__main__":

    def fitness_func(chromosome):
        """
        Sample fitness function.

        :param chromosome: A list representing a permutation of indices
        :return: An integer representing a fitness score
        """
        t1 = 1
        t2 = 1
        size = len(chromosome)
        f1 = [chromosome[i] - i for i in range(size)]
        f2 = [size - g for g in f1]
        f1.sort()
        f2.sort()
        for i in range(1, size):
            t1 += int(f1[i] == f1[i - 1])
            t2 += int(f2[i] == f2[i - 1])
        return t1 + t2

    optimizer = GeneticAlgorithm(1000, 0.8, 0.2, fitness_func)
    optimizer.find_optimum(8, 300)
    print(optimizer.population)



# PyZX - Python library for quantum circuit rewriting
#        and optimisation using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__version__ = "0.9.0"

from .graph.graph import Graph
from .circuit import Circuit, gates, id
from .linalg import Mat2
from .utils import settings, VertexType, EdgeType, original_colors, rgb_colors, grayscale_colors
from .quimb import to_quimb_tensor
from .drawing import *
from .simplify import *
from .optimize import *
from .extract import *
from .graph.jsonparser import *
from .tensor import *
from .local_search.simulated_annealing import anneal
from .local_search.genetic import GeneticOptimizer
from .circuit.qasmparser import qasm
from .circuit.sqasm import sqasm
from . import generate
from . import todd
from . import linalg
from . import extract
from .rewrite_rules import rules
from .rewrite_rules import hrules
from .rewrite_rules import basicrules
from .rewrite_rules import editor_actions
from . import optimize
from . import simplify
from . import hsimplify
from . import d3
from . import tikz
from . import simulate
from . import editor
from . import routing
from . import local_search
from .routing.parity_maps import CNOT_tracker

# some common scalars
from .graph.base import Scalar
ONE = Scalar()
SQRT_TWO = Scalar()
SQRT_TWO.add_power(1)
TWO = Scalar()
TWO.add_power(2)
SQRT_TWO_INV = Scalar()
SQRT_TWO_INV.add_power(-1)
TWO_INV = Scalar()
TWO_INV.add_power(-2)

if __name__ == '__main__':
    print("Please execute this as a module by running 'python -m pyzx'")



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import unittest
import sys
from types import ModuleType
from typing import Optional

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')
from pyzx_param.graph import Graph
from pyzx_param.utils import EdgeType, VertexType
from pyzx_param.quimb import to_quimb_tensor
from pyzx_param.simplify import full_reduce

np: Optional[ModuleType]
try:
    import numpy as np
    from pyzx_param.tensor import tensorfy, compare_tensors
except ImportError:
    np = None

try:
    import quimb as qu
    import quimb.tensor as qtn
except ImportError:
    qu = None


@unittest.skipUnless(np, "numpy needs to be installed for this to run")
@unittest.skipUnless(qu, "quimb needs to be installed for this to run")
class TestMapping(unittest.TestCase):
    def test_id_tensor(self):
        g = Graph()
        x = g.add_vertex(VertexType.BOUNDARY)
        y = g.add_vertex(VertexType.BOUNDARY)
        g.add_edge(g.edge(x, y), edgetype = EdgeType.SIMPLE)

        tn = to_quimb_tensor(g)
        self.assertTrue((tn & qtn.Tensor(data = [0, 1], inds = ("0",))
                            & qtn.Tensor(data = [0, 1], inds = ("1",)))
                        .contract(output_inds = ()) == 1)
        self.assertTrue((tn & qtn.Tensor(data = [1, 0], inds = ("0",))
                            & qtn.Tensor(data = [1, 0], inds = ("1",)))
                        .contract(output_inds = ()) == 1)

    def test_hadamard_tensor(self):
        g = Graph()
        x = g.add_vertex(VertexType.BOUNDARY)
        y = g.add_vertex(VertexType.BOUNDARY)
        g.add_edge(g.edge(x, y), edgetype = EdgeType.HADAMARD)

        tn = to_quimb_tensor(g)
        self.assertTrue(abs((tn & qtn.Tensor(data = [1, 0], inds = ("0",))
                            & qtn.Tensor(data = [1 / np.sqrt(2), 1 / np.sqrt(2)], inds = ("1",)))
                        .contract(output_inds = ()) - 1) < 1e-9)
        self.assertTrue(abs((tn & qtn.Tensor(data = [0, 1], inds = ("0",))
                            & qtn.Tensor(data = [1 / np.sqrt(2), -1 / np.sqrt(2)], inds = ("1",)))
                        .contract(output_inds = ()) - 1) < 1e-9)

    def test_xor_tensor(self):
        g = Graph()
        x = g.add_vertex(VertexType.BOUNDARY)
        y = g.add_vertex(VertexType.BOUNDARY)
        v = g.add_vertex(VertexType.Z)
        z = g.add_vertex(VertexType.BOUNDARY)

        g.add_edge(g.edge(x, v), edgetype = EdgeType.HADAMARD)
        g.add_edge(g.edge(y, v), edgetype = EdgeType.HADAMARD)
        g.add_edge(g.edge(v, z), edgetype = EdgeType.HADAMARD)
        tn = to_quimb_tensor(g)

        for x in range(2):
            for y in range(2):
                for z in range(2):
                    self.assertTrue(abs((tn &
                            qtn.Tensor(data = [1 - x, x], inds = ("0",)) &
                            qtn.Tensor(data = [1 - y, y], inds = ("1",)) &
                            qtn.Tensor(data = [1 - z, z], inds = ("3",))).contract(output_inds = ()) -
                            ((x ^ y) == z) / np.sqrt(2)) < 1e-9)

    def test_phases_tensor(self):
        # This diagram represents a 1-input 1-output Z-spider of phase pi/2,
        # but written using two Z-spiders of phases pi/6 and pi/3 that are
        # connected by a simple edge.
        g = Graph()
        x = g.add_vertex(VertexType.BOUNDARY)
        v = g.add_vertex(VertexType.Z, phase = 1. / 6.)
        w = g.add_vertex(VertexType.Z, phase = 1. / 3.)
        y = g.add_vertex(VertexType.BOUNDARY)

        g.add_edge(g.edge(x, v), edgetype = EdgeType.SIMPLE)
        g.add_edge(g.edge(v, w), edgetype = EdgeType.SIMPLE)
        g.add_edge(g.edge(w, y), edgetype = EdgeType.SIMPLE)
        tn = to_quimb_tensor(g)

        self.assertTrue(abs((tn & qtn.Tensor(data = [1, 0], inds = ("0",))
                            & qtn.Tensor(data = [1, 0], inds = ("3",)))
                        .contract(output_inds = ()) - 1) < 1e-9)
        self.assertTrue(abs((tn & qtn.Tensor(data = [0, 1], inds = ("0",))
                            & qtn.Tensor(data = [0, 1j], inds = ("3",)))
                        .contract(output_inds = ()) + 1) < 1e-9)

    def test_scalar(self):
        g = Graph()
        x = g.add_vertex(VertexType.Z, row = 0, phase = 1 / 2)
        y = g.add_vertex(VertexType.Z, row = 1, phase = 1 / 4)
        g.add_edge(g.edge(x, y), edgetype = EdgeType.SIMPLE)

        full_reduce(g)
        val = to_quimb_tensor(g).contract(output_inds = ())
        expected_val = 1 + np.exp(1j * np.pi * 3 / 4)
        self.assertTrue(abs(val - expected_val) < 1e-9)


if __name__ == '__main__':
    unittest.main()



# PyZX - Python library for quantum circuit rewriting
#       and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#   http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from fractions import Fraction
from typing import Optional, Tuple, Dict, Set, Any

from .base import BaseGraph

from ..utils import VertexType, EdgeType, FractionLike, FloatInt, vertex_is_zx_like, vertex_is_z_like, set_z_box_label, get_z_box_label, phase_to_s
import sympy as sym  # type: ignore[import-untyped]

class GraphS(BaseGraph[int,Tuple[int,int]]):
    """Purely Pythonic implementation of :class:`~graph.base.BaseGraph`."""
    backend = 'simple'

    #The documentation of what these methods do
    #can be found in base.BaseGraph
    def __init__(self) -> None:
        BaseGraph.__init__(self)
        self.graph: Dict[int,Dict[int,EdgeType]]   = dict()
        self._vindex: int                               = 0
        self.nedges: int                                = 0
        self.ty: Dict[int,VertexType]              = dict()
        self._phase: Dict[int, FractionLike]            = dict()
        self._phaseVars: Dict[int,Set[str]]             = dict()
        self._qindex: Dict[int, FloatInt]               = dict()
        self._maxq: FloatInt                            = -1
        self._rindex: Dict[int, FloatInt]               = dict()
        self._maxr: FloatInt                            = -1
        self._grounds: Set[int] = set()

        self._vdata: Dict[int,Any]                      = dict()
        self._edata: Dict[Tuple[int,int],Any] = dict()
        self._inputs: Tuple[int, ...]                   = tuple()
        self._outputs: Tuple[int, ...]                  = tuple()

    def clone(self) -> 'GraphS':
        cpy = GraphS()
        for v, d in self.graph.items():
            cpy.graph[v] = d.copy()
        cpy._vindex = self._vindex
        cpy.nedges = self.nedges
        cpy.ty = self.ty.copy()
        cpy._phase = self._phase.copy()
        cpy._phaseVars = self._phaseVars.copy()
        cpy._qindex = self._qindex.copy()
        cpy._maxq = self._maxq
        cpy._rindex = self._rindex.copy()
        cpy._maxr = self._maxr
        cpy._vdata = self._vdata.copy()
        cpy.scalar = self.scalar.copy()
        cpy._inputs = tuple(list(self._inputs))
        cpy._outputs = tuple(list(self._outputs))
        cpy.track_phases = self.track_phases
        cpy.phase_index = self.phase_index.copy()
        cpy.phase_master = self.phase_master
        cpy.phase_mult = self.phase_mult.copy()
        cpy.max_phase_index = self.max_phase_index
        return cpy

    def vindex(self): return self._vindex
    def depth(self):
        if self._rindex: self._maxr = max(self._rindex.values())
        else: self._maxr = -1
        return self._maxr
    def qubit_count(self):
        if self._qindex: self._maxq = max(self._qindex.values())
        else: self._maxq = -1
        return self._maxq + 1

    def inputs(self):
        return self._inputs

    def num_inputs(self):
        return len(self._inputs)

    def set_inputs(self, inputs):
        self._inputs = inputs

    def outputs(self):
        return self._outputs

    def num_outputs(self):
        return len(self._outputs)

    def set_outputs(self, outputs):
        self._outputs = outputs

    def add_vertices(self, amount):
        for i in range(self._vindex, self._vindex + amount):
            self.graph[i] = dict()
            self.ty[i] = VertexType.BOUNDARY
            self._phase[i] = 0
            self._phaseVars[i] = set()
        self._vindex += amount
        return range(self._vindex - amount, self._vindex)
    def add_vertex_indexed(self, v):
        """Adds a vertex that is guaranteed to have the chosen index (i.e. 'name').
        If the index isn't available, raises a ValueError.
        This method is used in the editor to support undo, which requires vertices
        to preserve their index."""
        if v in self.graph: raise ValueError("Vertex with this index already exists")
        if v >= self._vindex: self._vindex = v+1
        self.graph[v] = dict()
        self.ty[v] = VertexType.BOUNDARY
        self._phase[v] = 0
        self._phaseVars[v] = set()

    def add_edges(self, edge_pairs, edgetype=EdgeType.SIMPLE):
        for s,t in edge_pairs:
            self.nedges += 1
            self.graph[s][t] = edgetype
            self.graph[t][s] = edgetype
    
    def add_edge(self, edge_pair, edgetype=EdgeType.SIMPLE):
        s,t = edge_pair
        t1 = self.ty[s]
        t2 = self.ty[t]
        if s == t:
            if not vertex_is_zx_like(t1) or not vertex_is_zx_like(t2):
                raise ValueError(f'Unexpected vertex type, it should be either z or x because you are trying to add a self-loop')
            if edgetype==EdgeType.SIMPLE:
                return edge_pair
            elif edgetype==EdgeType.HADAMARD:
                self.add_to_phase(s, 1)
                return edge_pair
            else:
                raise ValueError(f'The edge you are adding is not an accepted type')
                
        if not t in self.graph[s]:
            self.nedges += 1
            self.graph[s][t] = edgetype
            self.graph[t][s] = edgetype
        else:
            if (vertex_is_zx_like(t1) and vertex_is_zx_like(t2)):
                et1 = self.graph[s][t]

                # set the roles of simple or hadamard edges, depending on whether the colours match
                if vertex_is_z_like(t1) == vertex_is_z_like(t2): # same colour
                    fuse, hopf = (EdgeType.SIMPLE, EdgeType.HADAMARD)
                else:
                    fuse, hopf = (EdgeType.HADAMARD, EdgeType.SIMPLE)

                # handle parallel edges for all possible combinations of fuse/hopf type edges
                if edgetype == fuse and et1 == fuse:
                    pass # no-op
                elif ((edgetype == fuse and et1 == hopf) or (edgetype == hopf and et1 == fuse)):
                    # ensure the remaining edge is 'fuse' type
                    self.set_edge_type((s,t), fuse)
                    # add a pi phase to one of the neighbours
                    if t1 == VertexType.Z_BOX:
                        set_z_box_label(self, s, get_z_box_label(self, s) * -1)
                    else:
                        self.add_to_phase(s, 1)
                    self.scalar.add_power(-1)
                elif edgetype == hopf and et1 == hopf:
                    # remove the edge (reducing mod 2)
                    self.remove_edge((s,t))
                    self.scalar.add_power(-2)
                else:
                    raise ValueError(f'Got unexpected edge types: {t1}, {t2}')
            else:
                raise ValueError(f'Attempted to add unreducible parallel edge {edge_pair}, types: {t1}, {t2}')


        return edge_pair

    def remove_vertices(self, vertices):
        for v in vertices:
            vs = list(self.graph[v])
            # remove all edges
            for v1 in vs:
                if v1 == v:
                    continue
                self.nedges -= 1
                del self.graph[v][v1]
                del self.graph[v1][v]
            # remove the vertex
            del self.graph[v]
            del self.ty[v]
            del self._phase[v]
            if v in self._inputs:
                self._inputs = tuple(u for u in self._inputs if u != v)
            if v in self._outputs:
                self._outputs = tuple(u for u in self._outputs if u != v)
            try: del self._qindex[v]
            except: pass
            try: del self._rindex[v]
            except: pass
            try: del self.phase_index[v]
            except: pass
            self._grounds.discard(v)
            self._vdata.pop(v,None)
        self._vindex = max(self.vertices(),default=0) + 1

    def remove_vertex(self, vertex):
        self.remove_vertices([vertex])

    def remove_edges(self, edges):
        for s,t in edges:
            if s == t:
                continue
            self.nedges -= 1
            del self.graph[s][t]
            del self.graph[t][s]
            self._edata.pop((s, t), None)

    def remove_edge(self, edge):
        self.remove_edges([edge])

    def num_vertices(self):
        return len(self.graph)

    def num_edges(self, s=None, t=None, et=None):
        if s is not None and t is not None:
            if self.connected(s, t):
                if et is not None:
                    if self.edge_type((s, t)) == et:
                        return 1
                    else:
                        return 0
                else:
                    return 1
            else:
                return 0
        elif s is not None:
            return self.vertex_degree(s)
        else:
            return len(list(self.edges()))

    def vertices(self):
        return self.graph.keys()

    def vertices_in_range(self, start, end):
        """Returns all vertices with index between start and end
        that only have neighbours whose indices are between start and end"""
        for v in self.graph.keys():
            if not start<v<end: continue
            if all(start<v2<end for v2 in self.graph[v]):
                yield v

    def edges(self, s=None, t=None):
        if s is not None and t is not None:
            if self.connected(s, t):
                yield (s,t) if s < t else (t,s)
        elif s is not None:
            for t in self.graph[s]:
                yield (s,t) if s < t else (t,s)
        else:
            for v0,adj in self.graph.items():
                for v1 in adj:
                    if v1 > v0: yield (v0,v1)

    def edges_in_range(self, start, end, safe=False):
        """like self.edges, but only returns edges that belong to vertices
        that are only directly connected to other vertices with
        index between start and end.
        If safe=True then it also checks that every neighbour is only connected to vertices with the right index"""
        if not safe:
            for v0,adj in self.graph.items():
                if not (start<v0<end): continue
                #verify that all neighbours are in range
                if all(start<v1<end for v1 in adj):
                    for v1 in adj:
                        if v1 > v0: yield (v0,v1)
        else:
            for v0,adj in self.graph.items():
                if not (start<v0<end): continue
                #verify that all neighbours are in range, and that each neighbour
                # is only connected to vertices that are also in range
                if all(start<v1<end for v1 in adj) and all(all(start<v2<end for v2 in self.graph[v1]) for v1 in adj):
                    for v1 in adj:
                        if v1 > v0:
                            yield (v0,v1)

    def edge(self, s, t):
        return (s,t) if s < t else (t,s)
    def edge_set(self):
        return set(self.edges())
    def edge_st(self, edge):
        return edge

    def neighbors(self, vertex):
        return self.graph[vertex].keys()

    def vertex_degree(self, vertex):
        return len(self.graph[vertex])

    def incident_edges(self, vertex):
        return [(vertex, v1) if v1 > vertex else (v1, vertex) for v1 in self.graph[vertex]]

    def connected(self,v1,v2):
        return v2 in self.graph[v1]

    def edge_type(self, e):
        v1,v2 = e
        try:
            return self.graph[v1][v2]
        except KeyError:
            return 0

    def set_edge_type(self, e, t):
        v1,v2 = e
        self.graph[v1][v2] = t
        self.graph[v2][v1] = t

    def type(self, vertex):
        return self.ty[vertex]
    def types(self):
        return self.ty
    def set_type(self, vertex, t):
        self.ty[vertex] = t

    def phase(self, vertex):
        return self._phase.get(vertex,Fraction(1))
    def phases(self):
        return self._phase
    def get_all_params(self):
        return self._phaseVars
    def get_params(self, vertex):
        return self._phaseVars[vertex]
    def set_params(self, vertex, params):
        self._phaseVars[vertex].clear()
        self._phaseVars[vertex] = self._phaseVars[vertex].union(params)
    def add_params(self, vertex, params):
        self._phaseVars[vertex] = self._phaseVars[vertex].union(params)
    def set_phase(self, vertex, phase, clearParams:Optional[bool]=True):
        if (isinstance(phase, str)): # if symbolic
            self._phaseVars[vertex].clear()
            self._phase[vertex] = 0
            self._phaseVars[vertex].add(phase)
        else:
            if (clearParams): self._phaseVars[vertex].clear()
            try:
                self._phase[vertex] = Fraction(phase) % 2
            except Exception:
                self._phase[vertex] = phase
    def add_to_phase(self, vertex, phase, params=None):
        if params is None:
            params = set()
        self._phaseVars[vertex] = self._phaseVars[vertex].symmetric_difference(params) # XOR
        old_phase = self._phase.get(vertex, Fraction(1))
        try:
            self._phase[vertex] = (old_phase + Fraction(phase)) % 2
        except Exception:
            self._phase[vertex] = old_phase + phase
    def toggle_var(self, vertex, strVar): # toggle a variable's coefficient
        if (strVar in self._phaseVars[vertex]): self._phaseVars[vertex].remove(strVar)
        else: self._phaseVars[vertex].add(strVar)
    def qubit(self, vertex):
        return self._qindex.get(vertex,-1)
    def qubits(self):
        return self._qindex
    def set_qubit(self, vertex, q):
        if q > self._maxq: self._maxq = q
        self._qindex[vertex] = q

    def row(self, vertex):
        return self._rindex.get(vertex, -1)
    def rows(self):
        return self._rindex
    def set_row(self, vertex, r):
        if r > self._maxr: self._maxr = r
        self._rindex[vertex] = r

    def is_ground(self, vertex):
        return vertex in self._grounds
    def grounds(self):
        return self._grounds
    def set_ground(self, vertex, flag=True):
        if flag:
            self._grounds.add(vertex)
        else:
            self._grounds.discard(vertex)

    def clear_vdata(self, vertex):
        if vertex in self._vdata:
            del self._vdata[vertex]
    def vdata_keys(self, vertex):
        return self._vdata.get(vertex, {}).keys()
    def vdata(self, vertex, key, default=None):
        if vertex in self._vdata:
            return self._vdata[vertex].get(key,default)
        else:
            return default
    def set_vdata(self, vertex, key, val):
        if vertex in self._vdata:
            self._vdata[vertex][key] = val
        else:
            self._vdata[vertex] = {key:val}

    def clear_edata(self, edge):
        self._edata.pop(edge, None)
    def edata_keys(self, edge):
        return self._edata.get(edge, {}).keys()
    def edata(self, edge, key, default=None):
        if edge in self._edata:
            return self._edata[edge].get(key, default)
        else:
            return default
    def set_edata(self, edge, key, val):
        if edge in self._edata:
            self._edata[edge][key] = val
        else:
            self._edata[edge] = {key: val}

    def get_phase_sym(self, vertex):
        expr = 0
        for param in self._phaseVars[vertex]:
            expr = expr + sym.Symbol(param)
        return expr

    def get_phase_str(self, vertex):
        strExprVars = str(self.get_phase_sym(vertex))
        strExprPlus = " + "
        strExprConst = phase_to_s(self.phase(vertex), self.type(vertex))
        if (strExprVars == "0"): strExprVars = ""
        if (strExprConst == "" or strExprVars == ""): strExprPlus = ""
        return strExprConst + strExprPlus + strExprVars



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .qasmparser import QASMParser
from ..rewrite_rules.rules import *
from ..simplify import simp
from ..graph.base import BaseGraph

__all__ = ['sqasm']

#TODO: Improve the type annotation of these functions

# versions of these rules which instruct the simplifier *not* to remove
# isolated vertices. n.b. remove_ids already does this, but this might change
# in the future...
def spider_nocheck(g: BaseGraph, ms: List) -> RewriteOutputType:
    etab,rem_v,rem_e,check = spider(g, ms)
    return (etab, rem_v, rem_e, False)

def remove_ids_nocheck(g: BaseGraph, ms: List) -> RewriteOutputType:
    etab,rem_v,rem_e,check = remove_ids(g, ms)
    return (etab, rem_v, rem_e, False)

def sqasm(s: str, simplify=True) -> BaseGraph:
    p = QASMParser()
    c = p.parse(s, strict=False)
    g = c.to_graph(zh=True)
    inputs = list(g.inputs())
    outputs = list(g.outputs())
    for r,sp in p.registers.items():
        if len(r) > 0 and r[0].isupper():
            for q in range(sp[0],sp[0]+sp[1]):
                if r[0] != 'Z':
                    v = inputs[q]
                    v1 = list(g.neighbors(v))
                    if len(v1) > 0 and g.type(v1[0]) != VertexType.BOUNDARY:
                        g.set_type(v, g.type(v1[0]))
                    else:
                        g.set_type(v, VertexType.Z)
                    inputs[q] = None
                    g.scalar.add_power(-1)

                if r[0] != 'A':
                    v = outputs[q]
                    v1 = list(g.neighbors(v))
                    if len(v1) > 0 and g.type(v1[0]) != VertexType.BOUNDARY:
                        g.set_type(v, g.type(v1[0]))
                    else:
                        g.set_type(v, VertexType.Z)
                    outputs[q] = None
                    g.scalar.add_power(-1)
        
    g.set_inputs(tuple(x for x in inputs if not x is None))
    g.set_outputs(tuple(x for x in outputs if not x is None))
    
    while simplify:
        i = simp(g, '', match_spider_parallel, spider_nocheck, quiet=True)
        i += simp(g, '', match_ids_parallel, remove_ids_nocheck, quiet=True)
        if i == 0: break
    
    g.pack_circuit_rows()
    return g



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides methods for converting ZX-graphs into numpy tensors
and using these tensors to test semantic equality of ZX-graphs.
This module is not meant as an efficient quantum simulator.
Due to the way the tensor is calculated it can only handle
circuits of small size before running out of memory on a regular machine.
Currently, it can reliably transform 9 qubit circuits into tensors.
If the ZX-diagram is not circuit-like, but instead has nodes with high degree,
it will run out of memory even sooner."""

__all__ = ['tensorfy', 'compare_tensors', 'compose_tensors',
            'adjoint', 'is_unitary','tensor_to_matrix',
            'find_scalar_correction']

import itertools
from math import pi, sqrt

from typing import Optional

from .symbolic import Poly


import numpy as np
np.set_printoptions(suppress=True)

# typing imports
from typing import TYPE_CHECKING, List, Dict, Union
from .utils import FractionLike, FloatInt, VertexType, EdgeType, get_z_box_label
if TYPE_CHECKING:
    from .graph.base import BaseGraph, VT, ET
    from .circuit import Circuit
TensorConvertible = Union[np.ndarray, 'Circuit', 'BaseGraph']

def Z_box_to_tensor(arity: int, parameter: complex) -> np.ndarray:
    m = np.zeros([2]*arity, dtype = complex)
    if arity == 0:
        m[()] = 1 + parameter
        return m
    m[(0,)*arity] = 1
    m[(1,)*arity] = parameter
    return m

def Z_to_tensor(arity: int, phase: float) -> np.ndarray:
    return Z_box_to_tensor(arity, np.exp(1j*phase))

def X_to_tensor(arity: int, phase: float) -> np.ndarray:
    if arity == 0:
      m = np.zeros([2]*arity, dtype = complex)
      m[()] = 1 + np.exp(1j*phase)
      return m
    m = np.ones(2**arity, dtype = complex)
    for i in range(2**arity):
        if bin(i).count("1")%2 == 0:
            m[i] += np.exp(1j*phase)
        else:
            m[i] -= np.exp(1j*phase)
    return np.power(np.sqrt(0.5),arity)*m.reshape([2]*arity)

def H_to_tensor(arity: int, phase: float) -> np.ndarray:
    m = np.ones(2**arity, dtype = complex)
    if phase != 0: m[-1] = np.exp(1j*phase)
    return m.reshape([2]*arity)

def W_to_tensor(arity: int) -> np.ndarray:
    m = np.zeros([2]*arity,dtype=complex)
    if arity == 0:
        return m
    for i in range(arity):
        index = [0,]*arity
        index[i] = 1
        m[tuple(index)] = 1
    return m

def pop_and_shift(verts, indices):
    res = [indices[v].pop() for v in verts if v in indices]
    for i in sorted(res,reverse=True):
        for w,l in indices.items():
            l2 = []
            for j in l:
                if j>i: l2.append(j-1)
                else: l2.append(j)
            indices[w] = l2
    return res

def tensorfy(g: 'BaseGraph[VT,ET]', preserve_scalar=True, strategy='naive', verbose=False) -> np.ndarray:
    """Takes in a Graph and outputs a multidimensional numpy array
        representing the linear map the ZX-diagram implements.
        Beware that quantum circuits take exponential memory to represent."""
    if g.is_hybrid():
        raise ValueError("Hybrid graphs are not supported.")
    if strategy == 'naive':
        return tensorfy_naive(g, preserve_scalar=preserve_scalar)
    elif strategy.startswith('rw-'):
        from .rank_width import tensorfy_rw
        return tensorfy_rw(g, strategy=strategy, preserve_scalar=preserve_scalar, verbose=verbose)
    else:
        raise ValueError('Unknown simulation strategy')

def tensorfy_naive(g: 'BaseGraph[VT,ET]', preserve_scalar:bool=True) -> np.ndarray:
    rows = g.rows()
    phases = g.phases()
    types = g.types()
    depth = g.depth()
    verts_row: Dict[FloatInt, List['VT']] = {}
    for v in g.vertices():
        r = rows[v]
        if r in verts_row: verts_row[r].append(v)
        else: verts_row[r] = [v]

    inputs = g.inputs()
    outputs = g.outputs()
    if not inputs and not outputs:
        if any(g.type(v)==VertexType.BOUNDARY for v in g.vertices()):
            raise ValueError("Diagram contains BOUNDARY-type vertices, but has no inputs or outputs set. Perhaps call g.auto_detect_io() first?")

    had = 1/sqrt(2)*np.array([[1,1],[1,-1]])
    id2 = np.identity(2)
    tensor = np.array(1.0,dtype='complex128')
    qubits = len(inputs)
    for i in range(qubits): tensor = np.tensordot(tensor,id2,axes=0)
    inputs = tuple(inputs)
    indices = {}
    for i, v in enumerate(inputs):
        indices[v] = [1 + 2*i]

    for i,r in enumerate(sorted(verts_row.keys())):
        for v in sorted(verts_row[r]):
            if types[v] == VertexType.DUMMY:
                continue
            neigh = list(itertools.chain.from_iterable(
                set(g.edge_st(e)) - {v} for e in g.incident_edges(v)
            ))
            self_loops = [e for e in g.incident_edges(v) if g.edge_s(e) == g.edge_t(e)]
            d = len(neigh) + len(self_loops) * 2
            if v in inputs:
                if types[v] != VertexType.BOUNDARY: raise ValueError("Wrong type for input:", v, types[v])
                continue # inputs already taken care of
            if v in outputs:
                if d != 1: raise ValueError("Weird output")
                if types[v] != VertexType.BOUNDARY: raise ValueError("Wrong type for output:",v, types[v])
                d += 1
                t = id2
            else:
                p = phases[v]
                if isinstance(p, Poly):
                    raise ValueError(f"Can't convert diagram with parameters to tensor: {str(p)}")
                phase = pi*p
                if types[v] == VertexType.Z:
                    t = Z_to_tensor(d,phase)
                elif types[v] == VertexType.X:
                    t = X_to_tensor(d,phase)
                elif types[v] == VertexType.H_BOX:
                    t = H_to_tensor(d,phase)
                elif types[v] == VertexType.W_INPUT or types[v] == VertexType.W_OUTPUT:
                    if phase != 0: raise ValueError("Phase on W node")
                    t = W_to_tensor(d)
                elif types[v] == VertexType.Z_BOX:
                    if phase != 0: raise ValueError("Phase on Z box")
                    label = get_z_box_label(g, v)
                    t = Z_box_to_tensor(d, label)
                else:
                    raise ValueError("Vertex %s has non-ZXH type but is not an input or output" % str(v))
            for sl in self_loops:
                if g.edge_type(sl) == EdgeType.HADAMARD:
                    t = np.tensordot(t,had)
                elif g.edge_type(sl) == EdgeType.SIMPLE:
                    t = np.trace(t)
                else:
                    raise NotImplementedError(f"Tensor contraction with {repr(sl)} self-loops is not implemented.")
            nn = list(filter(lambda n: rows[n]<r or (rows[n]==r and n<v), neigh)) # TODO: allow ordering on vertex indices?
            ety = {n:g.edge_type(g.edge(v,n)) for n in nn}
            nn.sort(key=lambda n: ety[n])
            for n in nn:
                if ety[n] == EdgeType.HADAMARD:
                    t = np.tensordot(t,had,(0,0)) # Hadamard edges are moved to the last index of t
            contr = pop_and_shift(nn,indices) #the last indices in contr correspond to hadamard contractions
            tensor = np.tensordot(tensor,t,axes=(contr,list(range(len(t.shape)-len(contr),len(t.shape)))))
            indices[v] = list(range(len(tensor.shape)-d+len(contr), len(tensor.shape)))

            if not preserve_scalar and i % 10 == 0:
                if np.abs(tensor).max() < 10**-6: # Values are becoming too small
                    tensor *= 10**4 # So scale all the numbers up
    perm = []
    for o in outputs:
        perm.append(indices[o][0])
    for i in range(len(inputs)):
        perm.append(i)

    tensor = np.transpose(tensor,perm)
    if preserve_scalar: tensor *= g.scalar.to_number()
    return tensor

def tensor_to_matrix(t: np.ndarray, inputs: int, outputs: int) -> np.ndarray:
    """Takes a tensor generated by ``tensorfy`` and turns it into a matrix.
    The ``inputs`` and ``outputs`` arguments specify the final shape of the matrix:
    2^(outputs) x 2^(inputs)"""
    rows = []
    for r in range(2**outputs):
        if outputs == 0:
            o = []
        else:
            o = [int(i) for i in bin(r)[2:].zfill(outputs)]
        row = []
        if inputs == 0:
            row.append(t[tuple(o)])
        else:
            for c in range(2**inputs):
                a = o.copy()
                a.extend([int(i) for i in bin(c)[2:].zfill(inputs)])
                row.append(t[tuple(a)])
        rows.append(row)
    return np.array(rows)

def compare_tensors(t1: TensorConvertible,t2: TensorConvertible, preserve_scalar: bool=False) -> bool:
    """Returns true if ``t1`` and ``t2`` represent equal tensors.
    When `preserve_scalar` is False (the default), equality is checked up to nonzero rescaling.

    Example: To check whether two ZX-graphs `g1` and `g2` are semantically the same you would do::

        compare_tensors(g1,g2) # True if g1 and g2 represent the same linear map up to nonzero scalar

    """
    from .circuit import Circuit

    if not isinstance(t1, np.ndarray):
        t1 = t1.to_tensor(preserve_scalar)
    if not isinstance(t2, np.ndarray):
        t2 = t2.to_tensor(preserve_scalar)
    if np.allclose(t1,t2): return True
    if preserve_scalar: return False # We do not check for equality up to scalar
    epsilon = 10**-14
    for i,a in enumerate(t1.flat):
        if abs(a)>epsilon:
            if abs(t2.flat[i])<epsilon: return False
            break
    else:
        raise ValueError("Tensor is too close to zero")
    return np.allclose(t1/a,t2/t2.flat[i])

def find_scalar_correction(t1: TensorConvertible, t2:TensorConvertible) -> complex:
    """Returns the complex number ``z`` such that ``t1 = z*t2``.

    Warning:
        This function assumes that ``compare_tensors(t1,t2,preserve_scalar=False)`` is True,
        i.e. that ``t1`` and ``t2`` indeed are equal up to global scalar.
        If they aren't, this function returns garbage.

    """
    if not isinstance(t1, np.ndarray):
        t1 = t1.to_tensor(preserve_scalar=True)
    if not isinstance(t2, np.ndarray):
        t2 = t2.to_tensor(preserve_scalar=True)

    epsilon = 10**-14
    for i,a in enumerate(t1.flat):
        if abs(a)>epsilon:
            if abs(t2.flat[i])<epsilon: return 0
            return a/t2.flat[i]

    return 0


def compose_tensors(t1: np.ndarray, t2: np.ndarray) -> np.ndarray:
    """Returns a tensor that is the result of composing the tensors together as if they
    were representing circuits::

        t1 = tensorfy(circ1)
        t2 = tensorfy(circ2)
        circ1.compose(circ2)
        t3 = tensorfy(circ1)
        t4 = compose_tensors(t1,t2)
        compare_tensors(t3,t4) # This is True

    """

    if len(t1.shape) != len(t2.shape):
        raise TypeError("Tensors represent circuits of different amount of qubits, "
                        "{!s} vs {!s}".format(len(t1.shape)//2,len(t2.shape)//2))
    q = len(t1.shape)//2
    contr2 = [q+i for i in range(q)]
    contr1 = [i for i in range(q)]
    t = np.tensordot(t1,t2,axes=(contr1,contr2))
    transp = []
    for i in range(q):
        transp.append(q+i)
    for i in range(q):
        transp.append(i)
    return np.transpose(t,transp)


def adjoint(t: np.ndarray) -> np.ndarray:
    """Returns the adjoint of the tensor as if it were representing
    a circuit::

        t = tensorfy(circ)
        tadj = tensorfy(circ.adjoint())
        compare_tensors(adjoint(t),tadj) # This is True

    """

    q = len(t.shape)//2
    transp = []
    for i in range(q):
        transp.append(q+i)
    for i in range(q):
        transp.append(i)
    return np.transpose(t.conjugate(),transp)


def is_unitary(g: 'BaseGraph') -> bool:
    """Returns whether the given ZX-graph is equal to a unitary (up to a number)."""
    from .generate import identity # Imported here to prevent circularity
    adj = g.adjoint()
    adj.compose(g)
    return compare_tensors(adj.to_tensor(), identity(len(g.inputs()),2).to_tensor(), False)



import random
import math
import numpy as np
import sys

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')

from pyzx_param.graph.scalar import DyadicNumber
from pyzx_param.simulate import MAGIC_GLOBAL, MAGIC_B60, MAGIC_B66, MAGIC_E6, MAGIC_O6, MAGIC_K6, MAGIC_PHI


def test_mul():
    d1 = DyadicNumber(1, 1, 0, 0, 1)
    d2 = DyadicNumber(1, 0, 1, 1, 0)
    d3 = d1 * d2
    assert np.isclose(d3.to_complex(), d1.to_complex() * d2.to_complex())

    d1 = DyadicNumber(0, 0, 1, 0, 1)
    d2 = DyadicNumber(0, 0, 1, 0, 1)
    d3 = d1 * d2
    assert np.isclose(d3.to_complex(), d1.to_complex() * d2.to_complex())


def test_mul_with_random_scalars():
    for _ in range(20):
        d1 = DyadicNumber(*[random.randint(-20, 20) for _ in range(5)])
        d2 = DyadicNumber(*[random.randint(-20, 20) for _ in range(5)])
        d3 = d1 * d2
        assert np.isclose(d3.to_complex(), d1.to_complex() * d2.to_complex())


def test_sqrt2():
    sqrt2 = DyadicNumber.sqrt2()
    assert np.isclose(sqrt2.to_complex(), np.sqrt(2))


def test_one():
    one = DyadicNumber.one()
    assert np.isclose(one.to_complex(), 1)


def test_conjugate():
    for _ in range(20):
        d1 = DyadicNumber(*[random.randint(-20, 20) for _ in range(5)])
        d1_conj = d1.conjugate()
        assert np.isclose(d1.to_complex(), d1_conj.to_complex().conjugate())


def test_magic_constants():
    sq2 = math.sqrt(2)
    MAGIC_GLOBAL2 = -(7 + 5 * sq2) / (2 + 2j)
    MAGIC_B602 = -16 + 12 * sq2
    MAGIC_B662 = 96 - 68 * sq2
    MAGIC_E62 = 10 - 7 * sq2
    MAGIC_O62 = -14 + 10 * sq2
    MAGIC_K62 = 7 - 5 * sq2
    MAGIC_PHI2 = 10 - 7 * sq2

    assert MAGIC_GLOBAL2 == MAGIC_GLOBAL.to_complex()
    assert MAGIC_B602 == MAGIC_B60.to_complex()
    assert MAGIC_B662 == MAGIC_B66.to_complex()
    assert MAGIC_E62 == MAGIC_E6.to_complex()
    assert MAGIC_O62 == MAGIC_O6.to_complex()
    assert MAGIC_K62 == MAGIC_K6.to_complex()
    assert MAGIC_PHI2 == MAGIC_PHI.to_complex()




# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional

from . import Circuit
from .gates import TargetMapper
from ..utils import EdgeType, VertexType, FloatInt, FractionLike
from ..graph import Graph
from ..graph.base import BaseGraph, VT, ET

def graph_to_circuit(g:BaseGraph[VT,ET], split_phases:bool=True) -> Circuit:
    inputs = g.inputs()
    qs = g.qubits()
    rs = g.rows()
    ty = g.types()
    phases = g.phases()
    rows: Dict[FloatInt,List[VT]] = {}

    c = Circuit(len(inputs))
    
    for v in g.vertices():
        if v in inputs: continue
        r = g.row(v)
        if r in rows: rows[r].append(v)
        else: rows[r] = [v]
    for r in sorted(rows.keys()):
        for v in rows[r]:
            q = qs[v]
            phase = phases[v]
            t = ty[v]
            neigh = [w for w in g.neighbors(v) if rs[w]<r]
            if len(neigh) != 1:
                raise TypeError("Graph doesn't seem circuit like: multiple parents")
            n = neigh[0]
            if qs[n] != q:
                raise TypeError("Graph doesn't seem circuit like: cross qubit connections")
            if g.edge_type(g.edge(n,v)) == EdgeType.HADAMARD:
                c.add_gate("HAD", q)
            if t == VertexType.BOUNDARY: #vertex is an output
                continue
            if phase!=0 and not split_phases:
                if t == VertexType.Z: c.add_gate("ZPhase", q, phase=phase)
                else: c.add_gate("XPhase", q, phase=phase)
            elif t == VertexType.Z and phase.denominator == 2:
                c.add_gate("S", q, adjoint=(phase.numerator==3))
            elif t == VertexType.Z and phase.denominator == 4:
                if phase.numerator in (1,7): c.add_gate("T", q, adjoint=(phase.numerator==7))
                if phase.numerator in (3,5):
                    c.add_gate("Z", q)
                    c.add_gate("T", q, adjoint=(phase.numerator==3))
            elif phase == 1:
                if t == VertexType.Z: c.add_gate("Z", q)
                else: c.add_gate("NOT", q)
            elif phase != 0:
                if t == VertexType.Z: c.add_gate("ZPhase", q, phase=phase)
                else: c.add_gate("XPhase", q, phase=phase)

            neigh = [w for w in g.neighbors(v) if rs[w]==r and w<v] # TODO: find a different way to do comparison of vertices
            for n in neigh:
                t2 = ty[n]
                q2 = qs[n]
                if t == t2:
                    if g.edge_type(g.edge(v,n)) != EdgeType.HADAMARD:
                        raise TypeError("Invalid vertical connection between vertices of the same type")
                    if t == VertexType.Z: c.add_gate("CZ", q2, q)
                    else: c.add_gate("CX", q2, q)
                else:
                    if g.edge_type(g.edge(v,n)) != EdgeType.SIMPLE:
                        raise TypeError("Invalid vertical connection between vertices of different type")
                    if t == VertexType.Z: c.add_gate("CNOT", q, q2)
                    else: c.add_gate("CNOT", q2, q)
    return c


def circuit_to_graph(c: Circuit, compress_rows:bool=True, backend:Optional[str]=None) -> BaseGraph[VT, ET]:
    """Turns the circuit into a ZX-Graph.
    If ``compress_rows`` is set, it tries to put single qubit gates on different qubits,
    on the same row."""
    g = Graph(backend)
    q_mapper: TargetMapper[VT] = TargetMapper()
    c_mapper: TargetMapper[VT] = TargetMapper()
    inputs = []
    outputs = []

    for i in range(c.qubits):
        v = g.add_vertex(VertexType.BOUNDARY,i,0)
        inputs.append(v)
        q_mapper.add_label(i, 1)
        q_mapper.set_prev_vertex(i, v)
    for i in range(c.bits):
        qubit = i+c.qubits
        v = g.add_vertex(VertexType.BOUNDARY, qubit, 0)
        inputs.append(v)
        q_mapper.add_label(qubit, 1)
        c_mapper.set_prev_vertex(qubit, v)

    for gate in c.gates:
        if gate.name == 'InitAncilla':
            l = gate.label # type: ignore
            try:
                q_mapper.add_label(l, q_mapper.next_row_or_default(l, q_mapper.max_row() - 1))
            except ValueError:
                raise ValueError("Ancilla label {} already in use".format(str(l)))
            v = g.add_vertex(VertexType.Z, q_mapper.to_qubit(l), q_mapper.next_row(l))
            q_mapper.set_prev_vertex(l, v)
            q_mapper.advance_next_row(l)
        elif gate.name == 'PostSelect':
            l = gate.label # type: ignore
            try:
                q = q_mapper.to_qubit(l)
                r = q_mapper.next_row(l)
                u = q_mapper.prev_vertex(l)
                q_mapper.set_next_row(l, r + 1)
                q_mapper.remove_label(l)
            except ValueError:
                raise ValueError("PostSelect label {} is not in use".format(str(l)))
            v = g.add_vertex(VertexType.Z, q, r)
            g.add_edge((u,v),EdgeType.SIMPLE)
        else:
            if not compress_rows: #or not isinstance(gate, (ZPhase, XPhase, HAD)):
                q_mapper.set_max_row(max(q_mapper.max_row(), c_mapper.max_row()))
                c_mapper.set_max_row(q_mapper.max_row())
                q_mapper.set_all_rows_to_max()
                c_mapper.set_all_rows_to_max()
            gate.to_graph(g, q_mapper, c_mapper)
            if not compress_rows: # or not isinstance(gate, (ZPhase, XPhase, HAD)):
                q_mapper.set_max_row(max(q_mapper.max_row(), c_mapper.max_row()))
                c_mapper.set_max_row(q_mapper.max_row())
                q_mapper.set_all_rows_to_max()
                c_mapper.set_all_rows_to_max()

    r = max(q_mapper.max_row(), c_mapper.max_row())
    for mapper in (q_mapper, c_mapper):
        for l in mapper.labels():
            o = mapper.to_qubit(l)
            v = g.add_vertex(VertexType.BOUNDARY, o, r)
            outputs.append(v)
            u = mapper.prev_vertex(l)
            g.add_edge((u,v))

    g.set_inputs(tuple(inputs))
    g.set_outputs(tuple(outputs))

    return g



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

from .base import BaseGraph
from .graph_s import GraphS
from .multigraph import Multigraph

try:
	import quizx # type: ignore
except ImportError:
	quizx = None

backends = { 'simple': True, 'multigraph': True, 'quizx-vec': False if quizx is None else True }

def Graph(backend:Optional[str]=None) -> BaseGraph:
	"""Returns an instance of an implementation of :class:`~pyzx.graph.base.BaseGraph`. 
	By default :class:`~pyzx.graph.graph_s.GraphS` is used. 
	Currently ``backend`` is allowed to be `simple` (for the default),
	or 'graph_tool' and 'igraph'.
	This method is the preferred way to instantiate a ZX-diagram in PyZX.

	Example:
		To construct an empty ZX-diagram, just write::

			g = zx.Graph()
		
	"""
	if backend is None: backend = 'simple'
	if backend not in backends:
		raise KeyError("Unavailable backend '{}'".format(backend))
	if backend == 'simple': return GraphS()
	if backend == 'multigraph': return Multigraph()
	if backend == 'graph_tool': 
		return GraphGT()
	if backend == 'igraph': return GraphIG()
	if backend == 'quizx-vec': return quizx.VecGraph()
	return GraphS()

Graph.from_json = GraphS.from_json # type: ignore
Graph.from_tikz = GraphS.from_tikz # type: ignore

try:
	import graph_tool.all as gt
	from .graph_gt import GraphGT
	backends['graph_tool'] = gt
except ImportError:
	pass
try:
	import igraph as ig
	from .graph_ig import GraphIG
	backends['igraph'] = ig 
except ImportError:
	pass



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from . import Circuit

def circuit_to_emoji(c: Circuit, compress_rows:bool=True) -> str:
    """Turns the circuit into a ZX-Graph.
    If ``compress_rows`` is set, it tries to put single qubit gates on different qubits,
    on the same row."""
    strings: List[List[str]] = [list() for _ in range(c.qubits)]
    #for i in range(c.qubits):
    #    strings[i].append(':Wire_lr:')

    for gate in c.gates:
        if not compress_rows:
            r = max([len(s) for s in strings])
            for s in strings: 
                s.extend([':W_:']*(r-len(s)))
        if not hasattr(gate, "to_emoji"): raise TypeError("Gate {} cannot be converted to emoji".format(str(gate)))
        gate.to_emoji(strings)
        if not compress_rows:
            r = max([len(s) for s in strings])
            for s in strings: 
                s.extend([':W_:']*(r-len(s)))

    r = max([len(s) for s in strings])
    for s in strings: 
        s.extend([':W_:']*(r-len(s)))

    return "\n".join(["".join(s) for s in strings])



# PyZX - Python library for quantum circuit rewriting
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['string_to_phase','to_graphml']

import json
import re
import ast
from fractions import Fraction
from typing import List, Dict, Tuple, Any, Optional, Callable, Union, TYPE_CHECKING

from pyzx_param.graph.multigraph import Multigraph

from ..utils import FractionLike, EdgeType, VertexType, phase_to_s
from .graph import Graph
from .scalar import Scalar
from .base import BaseGraph, VT, ET
from ..simplify import id_simp
from ..symbolic import parse, Poly, new_var
if TYPE_CHECKING:
    from .diff import GraphDiff
    from .multigraph import Multigraph


def string_to_phase(string: str, g: Union[BaseGraph,'GraphDiff']) -> Union[Fraction, Poly]:
    if not string:
        return Fraction(0)
    try:
        s = string.lower().replace(' ', '')
        s = s.replace('*', '')
        s = re.sub(r'\\?(pi|\u03c0)', '', s)
        if s == '': return Fraction(1)
        if s == '-': return Fraction(-1)
        if '.' in s or 'e' in s:
            return Fraction(float(s))
        elif '/' in s:
            a, b = s.split("/", 2)
            if not a:
                return Fraction(int(1), int(b)) # For some weird reason this int(1) is needed. See PR #234.
            if a == '-':
                a = '-1'
            return Fraction(int(a), int(b))
        else:
            return Fraction(int(s))
    except ValueError:
        def _new_var(name: str) -> Poly:
            if name not in g.variable_types:
                g.variable_types[name] = False
            return new_var(name, g.variable_types)
        try:
            return parse(string, _new_var)
        except Exception as e:
            raise ValueError(e)

def json_to_graph_old(js: Union[str,Dict[str,Any]], backend:Optional[str]=None) -> BaseGraph:
    """This method is deprecated. Use `json_to_graph` or `dict_to_graph` instead.

    Converts the json representation of a .qgraph Quantomatic graph into
    a pyzx graph. If JSON is given as a string, parse it first."""
    if isinstance(js, str):
        j = json.loads(js)
    else:
        j = js
    g = Graph(backend)
    g.variable_types = j.get('variable_types',{})

    names: Dict[str, Any] = {} # TODO: Any = VT
    hadamards: Dict[str, List[Any]] = {}
    for name,attr in j.get('node_vertices',{}).items():
        if ('data' in attr and 'type' in attr['data'] and attr['data']['type'] == "hadamard"
            and 'is_edge' in attr['data'] and attr['data']['is_edge'] == 'true'):
            hadamards[name] = []
            continue
        c = attr['annotation']['coord']
        q, r = -c[1], c[0]
        if q == int(q): q = int(q)
        if r == int(r): r = int(r)
        v = g.add_vertex(qubit=q, row=r)
        g.set_vdata(v,'name',name)
        names[name] = v
        if 'data' in attr:
            d = attr['data']
            if not 'type' in d or d['type'] == 'Z': g.set_type(v,VertexType.Z)
            elif d['type'] == 'X': g.set_type(v,VertexType.X)
            elif d['type'] == 'hadamard': g.set_type(v,VertexType.H_BOX)
            elif d['type'] == 'W_input': g.set_type(v,VertexType.W_INPUT)
            elif d['type'] == 'W_output': g.set_type(v,VertexType.W_OUTPUT)
            elif d['type'] == 'Z_box': g.set_type(v,VertexType.Z_BOX)
            else: raise TypeError("unsupported type '{}'".format(d['type']))
            if 'value' in d:
                g.set_phase(v,string_to_phase(d['value'],g))
            else:
                g.set_phase(v,Fraction(0,1))
            if d.get('ground', False):
                g.set_ground(v)
        else:
            g.set_type(v,VertexType.Z)
            g.set_phase(v,Fraction(0,1))
        for key, value in attr['annotation'].items():
            if key == 'coord':
                continue
            elif key == 'label':
                if type(value) != complex:
                    value = string_to_phase(value,g)
            g.set_vdata(v, key, value)

    inputs = {}
    outputs = {}

    for name,attr in j.get('wire_vertices',{}).items():
        ann = attr['annotation']
        c = ann['coord']
        q, r = -c[1], c[0]
        if q == int(q): q = int(q)
        if r == int(r): r = int(r)
        v = g.add_vertex(VertexType.BOUNDARY,q,r)
        g.set_vdata(v,'name',name)
        names[name] = v
        if "input" in ann: inputs[v] = ann["input"]
        if "output" in ann: outputs[v] = ann["output"]
        for key, value in attr['annotation'].items():
            if key in ('boundary','coord','input','output'):
                continue
            g.set_vdata(v, key, value)

    g.set_inputs(tuple(sorted(inputs.keys(),key=lambda v:inputs[v])))
    g.set_outputs(tuple(sorted(outputs.keys(),key=lambda v:outputs[v])))

    edges: Dict[Any, List[int]] = {} # TODO: Any = ET
    for edge in j.get('undir_edges',{}).values():
        n1, n2 = edge['src'], edge['tgt']
        if n1 in hadamards and n2 in hadamards: #Both
            v = g.add_vertex(VertexType.Z)
            name = "v"+str(len(names))
            g.set_vdata(v, 'name',name)
            names[name] = v
            hadamards[n1].append(v)
            hadamards[n2].append(v)
            continue
        if n1 in hadamards:
            hadamards[n1].append(names[n2])
            continue
        if n2 in hadamards:
            hadamards[n2].append(names[n1])
            continue
        if 'type' in edge and edge['type'] == 'w_io':
            g.add_edge((names[n1],names[n2]), EdgeType.W_IO)
            continue

        amount = edges.get((names[n1],names[n2]),[0,0])
        amount[0] += 1
        edges[(names[n1],names[n2])] = amount

    for l in hadamards.values():
        if len(l) != 2: raise TypeError("Can't parse graphs with irregular Hadamard nodes")
        e = tuple(l)
        amount = edges.get(e,[0,0])
        amount[1] += 1
        edges[e] = amount

    if "scalar" in j:
        g.scalar = Scalar.from_json(j["scalar"])
    g.add_edge_table(edges)

    return g

def graph_to_dict(g: BaseGraph[VT,ET], include_scalar: bool=True) -> Dict[str, Any]:
    """Converts a PyZX graph into Python dict for JSON output.
    If include_scalar is set to True (the default), then this includes the value
    of g.scalar with the json, which will also be loaded by the ``from_json`` method."""
    d = {
        "version": 2,
        "backend": g.backend,
        "variable_types": g.variable_types, # Potential source of error: this dictionary is mutable
    }
    if hasattr(g,'name'):
        d['name'] = g.name
    if include_scalar:
        d["scalar"] = g.scalar.to_dict()
    d['inputs'] = g.inputs()
    d['outputs'] = g.outputs()
    # Convert tuple keys to strings for JSON compatibility, replacing EdgeType with its integer value
    # This only applies to edata in Multigraph. For other classes, it returns str(k)
    def edata_key_to_str(k):
        if isinstance(k, tuple) and len(k) == 3 and hasattr(k[2], 'value'):
            return str((k[0], k[1], k[2].value))
        return str(k)
    d['edata'] = {edata_key_to_str(k): v for k, v in g._edata.items()}  # type: ignore[attr-defined]
    if g.backend == 'multigraph':
        d['auto_simplify'] = g.get_auto_simplify()

    verts = []
    for v in g.vertices():
        d_v = {
            'id': v,
            't': g.type(v),
            'pos': (round(g.row(v),3),round(g.qubit(v),3)),
        }
        if g.phase(v):
            d_v['phase'] = phase_to_s(g.phase(v))
        vdata_keys = g.vdata_keys(v)
        if vdata_keys:
            d_v['data'] = {k: g.vdata(v,k) for k in vdata_keys}
        if g.is_ground(v):
            d_v['is_ground'] = True
        verts.append(d_v)

    edges: List[Tuple[VT,VT,EdgeType]] = []
    if g.backend == 'multigraph':
        for e in g.edges():
            edges.append(e)  # type: ignore  # We know what we are doing, for multigraphs this has the right type.
    else:
        for e in g.edges():
            src,tgt = g.edge_st(e)
            et = g.edge_type(e)
            edges.append((src,tgt,et))

    d['vertices'] = verts
    d['edges']  = edges
    return d


def graph_to_dict_old(g: BaseGraph[VT,ET], include_scalar: bool=True) -> Dict[str, Any]:
    """This method is deprecated, and replaced by `graph_to_dict`.
    Converts a PyZX graph into Python dict for JSON output that is compatible with the Quantomatic format.
    If include_scalar is set to True (the default), then this includes the value
    of g.scalar with the json, which will also be loaded by the ``from_json`` method."""
    node_vs: Dict[str, Dict[str, Any]] = {}
    wire_vs: Dict[str, Dict[str, Any]] = {}
    edges: Dict[str, Dict[str, str]] = {}
    names: Dict[VT, str] = {}
    freenamesv = ["v"+str(i) for i in range(g.num_vertices()+g.num_edges())]
    freenamesb = ["b"+str(i) for i in range(g.num_vertices())]
    inputs = g.inputs()
    outputs = g.outputs()
    for v in g.vertices():
        t = g.type(v)
        coord = [round(g.row(v),3),round(-g.qubit(v),3)]
        name = g.vdata(v, 'name')
        if not name:
            if t == VertexType.BOUNDARY: name = freenamesb.pop(0)
            else: name = freenamesv.pop(0)
        else:
            try:
                freenamesb.remove(name) if t==VertexType.BOUNDARY else freenamesv.remove(name)
            except:
                pass

        names[v] = name
        if t == VertexType.BOUNDARY:
            wire_vs[name] = {"annotation":{"boundary":True,"coord":coord}}
            if v in inputs:
                wire_vs[name]["annotation"]["input"] = inputs.index(v)
            if v in outputs:
                wire_vs[name]["annotation"]["output"] = outputs.index(v)

            for key in g.vdata_keys(v):
                if key in wire_vs[name]["annotation"]:
                    continue
                wire_vs[name]["annotation"][key] = g.vdata(v, key)
        else:
            node_vs[name] = {"annotation": {"coord":coord},"data":{}}
            if t==VertexType.Z:
                node_vs[name]["data"]["type"] = "Z"
            elif t==VertexType.X:
                node_vs[name]["data"]["type"] = "X"
            elif t==VertexType.H_BOX:
                node_vs[name]["data"]["type"] = "hadamard"
                node_vs[name]["data"]["is_edge"] = "false"
            elif t==VertexType.W_INPUT:
                node_vs[name]["data"]["type"] = "W_input"
            elif t==VertexType.W_OUTPUT:
                node_vs[name]["data"]["type"] = "W_output"
            elif t==VertexType.Z_BOX:
                node_vs[name]["data"]["type"] = "Z_box"
                zbox_label = g.vdata(v, 'label', 1)
                if type(zbox_label) == Fraction:
                    zbox_label = phase_to_s(zbox_label)
                node_vs[name]["annotation"]["label"] = zbox_label
            else: raise Exception("Unkown vertex type "+ str(t))
            phase = phase_to_s(g.phase(v))
            if phase: node_vs[name]["data"]["value"] = phase
            if g.is_ground(v):
                node_vs[name]["data"]["ground"] = True
            if not node_vs[name]["data"]: del node_vs[name]["data"]
            for key in g.vdata_keys(v):
                if key in node_vs[name]["annotation"]:
                    continue
                node_vs[name]["annotation"][key] = g.vdata(v, key)

    i = 0
    for e in g.edges():
        src,tgt = g.edge_st(e)
        et = g.edge_type(e)
        if et == EdgeType.SIMPLE:
            edges["e"+ str(i)] = {"src": names[src],"tgt": names[tgt]}
            i += 1
        elif et == EdgeType.HADAMARD:
            x1,y1 = g.row(src), -g.qubit(src)
            x2,y2 = g.row(tgt), -g.qubit(tgt)
            hadname = freenamesv.pop(0)
            node_vs[hadname] = {"annotation": {"coord":[round((x1+x2)/2.0,3),round((y1+y2)/2.0,3)]},
                             "data": {"type": "hadamard","is_edge": "true"}}
            edges["e"+str(i)] = {"src": names[src],"tgt": hadname}
            i += 1
            edges["e"+str(i)] = {"src": names[tgt],"tgt": hadname}
            i += 1
        elif et == EdgeType.W_IO:
            edges["e"+str(i)] = {"src": names[src],"tgt": names[tgt], "type": "w_io"}
            i += 1
        else:
            raise TypeError("Edge of type 0")

    d: Dict[str,Any] = {
        "wire_vertices": wire_vs,
        "node_vertices": node_vs,
        "undir_edges": edges,
        "variable_types": g.variable_types,
    }
    if include_scalar:
        d["scalar"] = g.scalar.to_dict()

    return d

def graph_to_json(g: BaseGraph[VT,ET], include_scalar: bool=True) -> str:
    """Converts a PyZX graph into JSON output compatible with Quantomatic.
    If include_scalar is set to True (the default), then this includes the value
    of g.scalar with the json, which will also be loaded by the ``from_json`` method."""
    return json.dumps(graph_to_dict(g, include_scalar))

def dict_to_graph(d: Dict[str,Any], backend: Optional[str]=None) -> BaseGraph:
    """Converts a Python dict representation a graph produced by `graph_to_dict` into
    a pyzx Graph.
    If backend is given, it will be used as the backend for the graph, 
    otherwise the backend will be read from the dict description."""
    if not 'version' in d:
        # "Version is not specified in dictionary, will try to parse it as an older format")
        return json_to_graph_old(d, backend)
    else:
        if d['version'] != 2:
            raise ValueError("Unsupported version "+str(d['version']))
    if backend == None:
        backend = d.get('backend', None)
        if backend is None: raise ValueError("No backend specified in dictionary")
    
    g = Graph(backend)
    g.variable_types = d.get('variable_types',{})
    if g.backend == 'multigraph':
        if TYPE_CHECKING:
            assert isinstance(g, Multigraph)
        b = True if d.get('auto_simplify', True) in ('true', True) else False
        g.set_auto_simplify(b)
    for v_d in d['vertices']:
        pos = v_d['pos']
        v = v_d['id']
        g.add_vertex_indexed(v)
        g.set_type(v,v_d['t'])
        g.set_row(v,pos[0])
        g.set_qubit(v,pos[1])
        if 'phase' in v_d:
            g.set_phase(v,string_to_phase(v_d['phase'],g))
        if 'is_ground' in v_d and v_d['is_ground'] == True:
            g.set_ground(v)
        if 'data' in v_d:
            for k,val in v_d['data'].items():
                g.set_vdata(v,k,val)

    if 'edata' in d:
        g._edata = {ast.literal_eval(k): v for k, v in d['edata'].items()}  # type: ignore[attr-defined]

    for (s,t,et) in d['edges']:
        g.add_edge((s,t),et)

    return g

def json_to_graph(js: Union[str,Dict[str,Any]], backend:Optional[str]=None) -> BaseGraph:
    """Converts the json representation of a pyzx graph (as a string or dict) into
    a `Graph`. If JSON is given as a string, parse it first."""
    if isinstance(js, str):
        d = json.loads(js)
    else:
        d = js
    return dict_to_graph(d, backend)

def to_graphml(g: BaseGraph[VT,ET]) -> str:
    gml = """<?xml version="1.0" encoding="UTF-8" standalone="no"?>
<graphml xmlns="http://graphml.graphdrawing.org/xmlns">
    <key attr.name="type" attr.type="int" for="node" id="type">
        <default>1</default>
    </key>
    <key attr.name="phase" attr.type="string" for="node" id="phase">
        <default>0</default>
    </key>
    <key attr.name="edge type" attr.type="int" for="edge" id="etype">
        <default>1</default>
    </key>
    <key attr.name="x" attr.type="double" for="node" id="x">
        <default>0</default>
    </key>
    <key attr.name="y" attr.type="double" for="node" id="y">
        <default>0</default>
    </key>
    <graph edgedefault="undirected">
"""

    for v in g.vertices():
        gml += (
            (8*" " +
             """<node id="{!s}"><data key="type">{!s}</data><data key="phase">{!s}</data>"""+
             """<data key="x">{!s}</data><data key="y">{!s}</data></node>\n"""
            ).format(
                v, g.type(v), g.phase(v), g.row(v) * 100, g.qubit(v) * 100
            ))

    for e in g.edges():
        s,t = g.edge_st(e)
        gml += (
            (8*" " +
             """<edge id="{!s}_{!s}" source="{!s}" target="{!s}">"""+
             """<data key="etype">{!s}</data></edge>\n"""
            ).format(
                s, t, s, t, g.edge_type(e)
            ))
    gml += """
    </graph>
</graphml>
"""

    return gml





# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "cnots",
    "cliffords",
    "cliffordT",
    "cliffordTmeas",
    "identity",
    "CNOT_HAD_PHASE_circuit",
    "phase_poly",
    "phase_poly_approximate",
    "phase_poly_from_gadgets",
    "qft",
]

import random, types
from fractions import Fraction

from typing import Optional, List, Set, Union
from typing_extensions import Literal

import numpy as np
from pyzx_param.circuit.gates import CNOT, ZPhase
from pyzx_param.linalg import Mat2, MatLike

from pyzx_param.routing.parity_maps import CNOT_tracker, Parity
from pyzx_param.routing.phase_poly import PhasePoly, mat22partition

from .utils import EdgeType, VertexType, FloatInt, FractionLike, vertex_is_w, set_z_box_label
from .graph import Graph
from .graph.base import BaseGraph
from .circuit import Circuit

# TODO: This file needs some cleanup, as probably all functions should just return a Circuit instead of a Graph


def identity(qubits: int, depth: FloatInt=1,backend:Optional[str]=None) -> BaseGraph:
    """Generates a :func:`pyzx.graph.Graph` representing an identity circuit.

    Args:
        qubits: number of qubits (i.e. parallel lines of the Graph)
        depth: at which row the output vertices should be placed
        backend: the backend to use for the output graph
    """
    g = Graph(backend)
    inputs = []
    outputs = []
    for i in range(qubits):
        v = g.add_vertex(VertexType.BOUNDARY,i,0)
        w = g.add_vertex(VertexType.BOUNDARY,i,depth)
        inputs.append(v)
        outputs.append(w)
        g.add_edge((v,w))
    g.set_inputs(tuple(inputs))
    g.set_outputs(tuple(outputs))

    return g

def spider(
    typ:Union[Literal["Z"], Literal["X"], Literal["H"], Literal["W"], Literal["ZBox"], VertexType],
    inputs: int,
    outputs: int,
    phase:Optional[Union[FractionLike, complex]]=None,
    ) -> BaseGraph:
    """Returns a Graph containing a single spider of the specified type 
    and with the specified number of inputs and outputs."""
    if typ == "Z": typ = VertexType.Z
    elif typ == "X": typ = VertexType.X
    elif typ == "H": typ = VertexType.H_BOX
    elif typ == "W": typ = VertexType.W_OUTPUT
    elif typ == "ZBox": typ = VertexType.Z_BOX
    else:
        if not isinstance(typ, int):
            raise TypeError("Wrong type for spider type: " + str(typ))
    if phase is None:
        phase = 1 if typ == VertexType.Z_BOX else 0
    g = Graph()
    if vertex_is_w(typ):
        if inputs != 1:
            raise ValueError("Wrong number of inputs for W node: " + str(inputs))
        v_in = g.add_vertex(VertexType.W_INPUT, (outputs-1)/2, 0.8)
        v_out = g.add_vertex(VertexType.W_OUTPUT, (outputs-1)/2, 1)
        g.add_edge((v_in, v_out), EdgeType.W_IO)
    elif typ == VertexType.Z_BOX:
        v_in = g.add_vertex(typ, (inputs-1)/2, 1)
        set_z_box_label(g, v_in, phase)
        v_out = v_in
    else:
        assert isinstance(phase, (int, Fraction))
        v_in = g.add_vertex(typ, (inputs-1)/2, 1, phase)
        v_out = v_in
    inp = []
    outp = []
    for i in range(inputs):
        v = g.add_vertex(VertexType.BOUNDARY,i,0)
        inp.append(v)
    for i in range(outputs):
        v = g.add_vertex(VertexType.BOUNDARY,i,2)
        outp.append(v)
    for w in inp:
        g.add_edge((v_in, w))
    for w in outp:
        g.add_edge((v_out, w))

    g.set_inputs(tuple(inp))
    g.set_outputs(tuple(outp))

    return g


def CNOT_HAD_PHASE_circuit(
        qubits: int,
        depth: int,
        p_had: float = 0.2,
        p_t: float = 0.2,
        clifford:bool=False,
        seed:Optional[int]=None
        ) -> Circuit:
    """Construct a Circuit consisting of CNOT, HAD and phase gates.
    The default phase gate is the T gate, but if ``clifford=True``\\ , then
    this is replaced by the S gate.

    Args:
        qubits: number of qubits of the circuit
        depth: number of gates in the circuit
        p_had: probability that each gate is a Hadamard gate
        p_t: probability that each gate is a T gate (or if ``clifford`` is set, S gate)
        clifford: when set to True, the phase gates are S gates instead of T gates.

    Returns:
        A random circuit consisting of Hadamards, CNOT gates and phase gates.

    """
    rand: Union[random.Random, types.ModuleType]
    if (seed is not None): rand = random.Random(seed)
    else: rand = random
    p_cnot = 1-p_had-p_t
    c = Circuit(qubits)
    for _ in range(depth):
        r = rand.random()
        if r > 1-p_had:
            c.add_gate("HAD",rand.randrange(qubits))
        elif r > 1-p_had-p_t:
            if not clifford: c.add_gate("T",rand.randrange(qubits))
            else: c.add_gate("S",rand.randrange(qubits))
        else:
            tgt = rand.randrange(qubits)
            while True:
                ctrl = rand.randrange(qubits)
                if ctrl!=tgt: break
            c.add_gate("CNOT",tgt,ctrl)
    return c


def cnots(qubits: int, depth: int, backend:Optional[str]=None, seed:Optional[int]=None) -> BaseGraph:
    """Generates a circuit consisting of randomly placed CNOT gates.
    
    Args:
    qubits: Amount of qubits in circuit
    depth: Depth of circuit
    backend: When given, should be one of the possible :ref:`graph_api` backends.
    
    Returns:
        Instance of graph of the given backend
    """
    # initialise and add input row

    q: List[int] = list(range(qubits))                          # qubit index, initialised with input
    r: int = 1                                                  # current rank
    ty: List[VertexType] = [VertexType.BOUNDARY] * qubits  # types of vertices
    qs: List[int] = list(range(qubits))                         # tracks qubit indices of vertices
    rs: List[int] = [0] * qubits                                # tracks rank of vertices
    v = qubits                                                  # next vertex to add
    es = []                                                     # edges to add

    # initial row of Z
    for i in range(qubits):
        es.append((q[i], v))
        q[i] = v
        rs.append(r)
        qs.append(i)
        ty.append(VertexType.Z)
        v += 1
    r += 1

    # random CNOTs
    rand: Union[random.Random, types.ModuleType]
    if (seed is not None): rand = random.Random(seed)
    else: rand = random
    for i in range(depth):
        c = rand.randint(0, qubits-1)
        t = rand.randint(0, qubits-2)
        if t >= c: t += 1
        es += [(q[c], v), (q[t], v+1), (v, v+1)]
        q[c] = v
        q[t] = v+1
        rs += [r,r]
        qs += [c,t]
        ty += [VertexType.Z, VertexType.X]
        v += 2
        r += 1

    # final row of Z
    for i in range(qubits):
        es.append((q[i], v))
        q[i] = v
        rs.append(r)
        qs.append(i)
        ty.append(VertexType.Z)
        v += 1
    r += 1

    # outputs
    qs += list(range(qubits))
    rs += [r] * qubits
    ty += [VertexType.BOUNDARY] * qubits
    es += [(q[i], v+i) for i in range(qubits)]
    v += qubits

    g = Graph(backend)

    for i in range(v):
        g.add_vertex(ty[i],qs[i],rs[i])

    g.add_edges(es)

    inputs = []
    outputs = []
    for i in range(qubits):
        inputs.append(i)
        outputs.append(v-i-1)

    g.set_inputs(tuple(inputs))
    g.set_outputs(tuple(outputs))

    g.scalar.add_power(depth)
    return g

def accept(p: float, rand:Union[random.Random,types.ModuleType]=random) -> bool:
    return p>rand.random()

def random_phase(add_t: bool, rand:Union[random.Random,types.ModuleType]=random) -> Fraction:
    if add_t: return Fraction(rand.randint(1,8),4)
    return Fraction(rand.randint(1,4),2)

def cliffordTmeas(
        qubits: int, 
        depth: int, 
        p_t:Optional[float]=None, 
        p_s:Optional[float]=None, 
        p_hsh:Optional[float]=None, 
        p_cnot:Optional[float]=None, 
        p_meas:Optional[float]=None, 
        backend:Optional[str]=None,
        seed:Optional[int]=None
        ) -> BaseGraph:
    """Generates a circuit consisting of randomly placed Clifford+T gates. Optionally, take
    probabilities of adding T, S, HSH, CNOT, and measurements.
    If probabilities for only a subset of gates is given, any remaining probability will be
    uniformly distributed among the remaining gates.

    :param qubits: Amount of qubits in circuit.
    :param depth: Depth of circuit.
    :param p_t: Probability that each gate is a T-gate.
    :param p_s: Probability that each gate is a S-gate.
    :param p_hsh: Probability that each gate is a HSH-gate.
    :param p_cnot: Probability that each gate is a CNOT-gate.
    :param p_meas: Probability that each gate is a measurement.
    :param backend: When given, should be one of the possible :ref:`graph_api` backends.
    :rtype: Instance of graph of the given backend.
    """
    g = Graph(backend)
    qs = list(range(qubits))  # tracks qubit indices of vertices
    v = 0                     # next vertex to add
    r = 0                     # current row
    
    rand: Union[random.Random, types.ModuleType]
    if (seed is not None): rand = random.Random(seed)
    else: rand = random

    num = 0.0
    rest = 1.0
    if p_t is None: num += 1.0
    else: rest -= p_t
    if p_s is None: num += 1.0
    else: rest -= p_s
    if p_hsh is None: num += 1.0
    else: rest -= p_hsh
    if p_cnot is None: num += 1.0 if qubits > 1 else 0.0
    else: rest -= p_cnot
    if p_meas is None: num += 1.0
    else: rest -= p_meas

    if rest < 0: raise ValueError("Probabilities are >1.")

    if p_t is None: p_t = rest / num
    if p_s is None: p_s = rest / num
    if p_hsh is None: p_hsh = rest / num
    if p_cnot is None: p_cnot = rest / num if qubits > 1 else 0.0
    if p_meas is None: p_meas = rest / num

    if p_cnot > 0 and qubits <= 1: raise ValueError("Cannot have p_cnot > 0 with a single qubit.")

    #p_s = (1 - p_t) / 3.0
    #p_hsh = (1 - p_t) / 3.0
    #p_cnot = (1 - p_t) / 3.0

    inputs = []
    outputs = []

    for i in range(qubits):
        g.add_vertex(VertexType.BOUNDARY,i,r)
        inputs.append(v)
        v += 1
    r += 1

    for i in range(qubits):
        g.add_vertex(VertexType.Z,i,r)
        g.add_edge((qs[i], v))
        qs[i] = v
        v += 1
    r += 1

    for i in range(2, depth+2):
        p = rand.random()
        q0 = rand.randrange(qubits)

        g.add_vertex(VertexType.Z,q0,r)
        g.add_edge((qs[q0], v))
        qs[q0] = v
        v += 1
        r += 1

        if p > 1 - p_cnot:
            # apply CNOT gate
            q1 = rand.randrange(qubits-1)
            if q1 >= q0: q1 += 1

            g.add_vertex(VertexType.X,q1,r-1)
            g.add_edge((qs[q1], v))
            g.add_edge((v-1,v))
            g.scalar.add_power(1)
            qs[q1] = v
            v += 1
        elif p > 1 - p_cnot - p_hsh:
            # apply HSH gate
            g.set_type(v-1, VertexType.X)
            g.set_phase(v-1, Fraction(1,2))
        elif p > 1 - p_cnot - p_hsh - p_s:
            # apply S gate
            g.set_phase(v-1, Fraction(1,2))
        elif p > 1 - p_cnot - p_hsh - p_s - p_meas:
            # apply a measurement
            g.set_ground(v-1)
        else:
            # apply T gate
            g.set_phase(v-1, Fraction(1,4))

    for i in range(qubits):
        g.add_vertex(VertexType.Z,i,r)
        g.add_edge((qs[i], v))
        qs[i] = v
        v += 1
    r += 1

    for i in range(qubits):
        g.add_vertex(VertexType.BOUNDARY,i,r)
        g.add_edge((qs[i], v))
        outputs.append(v)
        v += 1

    g.set_inputs(tuple(inputs))
    g.set_outputs(tuple(outputs))

    return g

def cliffordT(
        qubits: int,
        depth: int,
        p_t:Optional[float]=None,
        p_s:Optional[float]=None,
        p_hsh:Optional[float]=None,
        p_cnot:Optional[float]=None,
        backend:Optional[str]=None,
        seed:Optional[int]=None
        ) -> BaseGraph:
    """Generates a circuit consisting of randomly placed Clifford+T gates. Optionally, take
    probabilities of adding T, S, HSH, and CNOT. If probabilities for only a subset of gates
    is given, any remaining probability will be uniformly distributed among the remaining
    gates.

    :param qubits: Amount of qubits in circuit.
    :param depth: Depth of circuit.
    :param p_t: Probability that each gate is a T-gate.
    :param p_s: Probability that each gate is a S-gate.
    :param p_hsh: Probability that each gate is a HSH-gate.
    :param p_cnot: Probability that each gate is a CNOT-gate.
    :param backend: When given, should be one of the possible :ref:`graph_api` backends.
    :rtype: Instance of graph of the given backend.
    """
    return cliffordTmeas(qubits, depth, p_t, p_s, p_hsh, p_cnot, 0, backend, seed)

def cliffords(
        qubits: int, 
        depth: int, 
        no_hadamard:bool=False,
        t_gates:bool=False,
        backend:Optional[str]=None,
        seed:Optional[int]=None
        ):
    """Generates a circuit consisting of randomly placed Clifford gates.
    Uses a different approach to generating Clifford circuits then :func:`cliffordT`.

    :param qubits: Amount of qubits in circuit.
    :param depth: Depth of circuit.
    :param no_hadamard: Whether hadamard edges are allowed to be placed.
    :param backend: When given, should be one of the possible :ref:`graph_api` backends.
    :rtype: Instance of graph of the given backend.
    """

    #randomness parameters
    p_two_qubit = 0.4 #whether to add a edge between two qubits
    p_cnot = 1 #0.4 # whether to CNOT or to CZ
    p_phase = 0.6 #probability of adding a phase to a node
    p_had = 0.2 # probability of adding a hadamard on a qubit

    # initialise and add input row

    q = list(range(qubits))   # qubit index, initialised with input
    r = 1                     # current rank
    ty: List[VertexType] = [VertexType.BOUNDARY] * qubits         # types of vertices
    qs = list(range(qubits))  # tracks qubit indices of vertices
    rs = [0] * qubits         # tracks rank of vertices
    v = qubits                # next vertex to add
    es1 = [] # normal edges to add
    es2 = [] # hadamard edges to add
    phases = {}

    # initial row of Z
    for i in range(qubits):
        es1.append((q[i], v))
        q[i] = v
        rs.append(r)
        qs.append(i)
        ty.append(VertexType.Z)
        v += 1
    r += 1

    # random gates
    rand: Union[random.Random, types.ModuleType]
    if (seed is not None): rand = random.Random(seed)
    else: rand = random
    for i in range(depth):
        c = rand.randint(0, qubits-1)
        t = rand.randint(0, qubits-2)
        if t >= c: t += 1
        if accept(p_two_qubit,rand):
            if no_hadamard or accept(p_cnot,rand): 
                es1.append((v, v+1))
                ty += [VertexType.Z, VertexType.X]
            else: 
                es2.append((v,v+1))
                typ: VertexType = rand.choice([VertexType.Z, VertexType.X])
                ty += [typ, typ]
            if accept(p_phase,rand): phases[v] = random_phase(t_gates,rand)
            if accept(p_phase,rand): phases[v+1] = random_phase(t_gates,rand)
        else:
            phases[v] = random_phase(t_gates,rand)
            phases[v+1] = random_phase(t_gates,rand)
            ty += [VertexType.Z, VertexType.X]
        
        if not no_hadamard and accept(p_had,rand): es2.append((q[c],v))
        else: es1.append((q[c],v))
        if not no_hadamard and accept(p_had,rand): es2.append((q[t],v+1))
        else: es1.append((q[t],v+1))

        q[c] = v
        q[t] = v+1
        
        rs += [r,r]
        qs += [c,t]
        v += 2
        r += 1

    # final row of Z
    for i in range(qubits):
        es1.append((q[i], v))
        q[i] = v
        rs.append(r)
        qs.append(i)
        ty.append(VertexType.Z)
        v += 1
    r += 1

    # outputs
    qs += list(range(qubits))
    rs += [r] * qubits
    ty += [VertexType.BOUNDARY] * qubits
    es1 += [(q[i], v+i) for i in range(qubits)]
    v += qubits

    g = Graph(backend)
    

    for i in range(v):
        g.add_vertex(ty[i],qs[i], rs[i])
    for w, phase in phases.items():
        g.set_phase(w,phase)

    g.add_edges(es1, EdgeType.SIMPLE)
    g.add_edges(es2, EdgeType.HADAMARD)

    inputs = []
    outputs = []

    for i in range(qubits):
        inputs.append(i)
        outputs.append(v-qubits+i)

    g.set_inputs(tuple(inputs))
    g.set_outputs(tuple(outputs))

    return g



def circuit_identity_phasepoly() -> Circuit:
    """Returns a 4-qubit circuit that is equal to the identity, provable by phase polynomial reductions."""
    c = Circuit(4)
    c.add_gate("ParityPhase",Fraction(1,4),0,1,2,3)
    for i in range(4):
        c.add_gate("ZPhase",i, Fraction(3,4))
        for j in range(i+1,4):
            c.add_gate("ParityPhase", Fraction(5,4),i,j)
            for k in range(j+1,4):
                c.add_gate("ParityPhase", Fraction(7,4),i,j,k)

    return c


def circuit_identity_commuting_controls(alpha:Fraction,beta:Fraction) -> Circuit:
    """Returns the circuit UVU*V* where U=NCZ(2beta) and V=CX(2alpha).
    Since these operations commute this circuit is equal to the identity.
    See page 13 of https://arxiv.org/pdf/1705.11151.pdf for more details."""
    cb = Circuit(2) 
    cb.add_gate("NOT",0)
    cb.add_gate("ZPhase",0,beta)
    cb.add_gate("ZPhase",1,beta)
    cb.add_gate("ParityPhase", -beta, 0, 1)
    cb.add_gate("NOT",0)

    ca = Circuit(2)
    ca.add_gate("ZPhase", 1, Fraction(3,2))
    ca.add_gate("XPhase", 1, Fraction(3,2))
    ca.add_gate("ZPhase",0,alpha)
    ca.add_gate("ZPhase",1,alpha)
    ca.add_gate("ParityPhase", -alpha, 0, 1)
    ca.add_gate("XPhase", 1, Fraction(1,2))
    ca.add_gate("ZPhase", 1, Fraction(1,2))

    c = cb.copy()
    c.add_circuit(ca)
    c.add_circuit(cb.adjoint())
    c.add_circuit(ca.adjoint())
    
    return c


def circuit_identity_two_qubit1() -> Circuit:
    """This returns the first nontrivial circuit identity from Selinger & Bian.
    See https://www.mathstat.dal.ca/~xbian/talks/slide_cliffordt2.pdf"""
    c = Circuit(2)
    c.add_gate("NOT",0)
    c.add_gate("T",1,adjoint=True)
    c.add_gate("S",1,adjoint=True)
    c.add_gate("HAD",1)
    c.add_gate("T",1,adjoint=True)
    c.add_gate("CNOT",0,1)
    c.add_gate("NOT",0)
    c.add_gate("T",1)
    c.add_gate("HAD",1)
    c.add_gate("S",1)
    c.add_gate("T",1)
    c.add_gate("CNOT",0,1)
    c.add_circuit(c)
    return c

def circuit_identity_two_qubit2() -> Circuit:
    """This returns the second nontrivial circuit identity from Selinger & Bian.
    See https://www.mathstat.dal.ca/~xbian/talks/slide_cliffordt2.pdf"""
    c = Circuit(2)
    c.add_gate("CNOT",0,1)
    c.add_gate("NOT",0)
    c.add_gate("T",1)
    c.add_gate("HAD",1)
    c.add_gate("T",1)
    c.add_gate("HAD",1)
    c.add_gate("T",1,adjoint=True)
    c.add_gate("CNOT",0,1)
    c.add_gate("NOT",0)
    c.add_gate("T",1)
    c.add_gate("HAD",1)
    c.add_gate("T",1,adjoint=True)
    c.add_gate("HAD",1)
    c.add_gate("T",1,adjoint=True)
    c.add_circuit(c)
    return c



def phase_poly(n_qubits: int, n_phase_layers: int, cnots_per_layer: int) -> Circuit:
    """
    Create a random phase polynomial circuit.

    :param n_qubits: Number of qubits in the circuit.
    :param n_phase_layers: Number of layers of phase gates.
    :param cnots_per_layer: Number of CNOTs in each layer.
    :return: A random phase polynomial circuit.
    """
    c = CNOT_tracker(n_qubits)
    for _ in range(n_phase_layers):
        build_random_parity_map(n_qubits, cnots_per_layer, circuit=c)
        for i in range(n_qubits):
            phase = Fraction(
                np.random.choice([1, -1]), int(np.random.choice([1, 2, 4]))
            )
            c.add_gate(ZPhase(target=i, phase=phase))
    return c


def phase_poly_approximate(n_qubits: int, n_CNOTs: int, n_phases: int) -> Circuit:
    """
    Create a random phase polynomial circuit with an exact number of CNOT gates.

    :param n_qubits: Number of qubits in the circuit.
    :param n_CNOTs: Number of CNOTs in the circuit.
    :param n_phases: Target of phase gates in the circuit. The actual number of phase gates may be slightly different.
    :return: A random phase polynomial circuit.
    """
    c = CNOT_tracker(n_qubits)
    cnot_count = 0
    p = n_phases / (n_CNOTs + n_phases)
    while cnot_count < n_CNOTs:
        target = np.random.randint(n_qubits)
        if np.random.rand() < p:
            phase = np.random.choice([1, -1]) * Fraction(
                1, int(np.random.choice([1, 2, 4]))
            )
            c.add_gate(ZPhase(target=target, phase=phase))
        else:
            control = np.random.choice([i for i in range(n_qubits) if i != target])
            c.add_gate(CNOT(control, target))
            cnot_count += 1
    return c


def phase_poly_from_gadgets(n_qubits: int, n_gadgets: int) -> Circuit:
    """
    Create a random phase polynomial circuit from a set of phase gadgets.

    :param n_qubits: Number of qubits in the circuit.
    :param n_gadgets: Number of phase gadgets to generate.
    :return: A random phase polynomial circuit.
    """
    parities: Set[Parity] = set()
    if n_gadgets > 2**n_qubits:
        n_gadgets = n_qubits ^ 3
    if n_qubits < 26:
        for integer in np.random.choice(
            2**n_qubits - 1, replace=False, size=n_gadgets
        ):
            parities.add(Parity(integer + 1, n_qubits))
    elif n_qubits < 64:
        while len(parities) < n_gadgets:
            parities.add(Parity(np.random.randint(1, 2**n_qubits), n_qubits))
    else:
        while len(parities) < n_gadgets:
            par: Parity = Parity([])
            while not par.count():
                par = Parity(
                    np.random.choice([False, True], n_qubits, replace=True), n_qubits
                )
            parities.add(par)
    zphase_dict = {p: Fraction(1, 4) for p in parities}
    out_parities = mat22partition(Mat2.id(n_qubits))
    phase_poly = PhasePoly(zphase_dict, out_parities)
    return phase_poly.rec_gray_synth("gauss", architecture=None)[0]


def build_random_parity_map(qubits: int, n_cnots: int, circuit=None) -> MatLike:
    """
    Builds a random parity map.

    :param qubits: The number of qubits that participate in the parity map
    :param n_cnots: The number of CNOTs in the parity map
    :param circuit: A (list of) circuit object(s) that implements a row_add() method to add the generated CNOT gates [optional]
    :return: a 2D numpy array that represents the parity map.
    """
    if circuit is None:
        circuit = []
    if not isinstance(circuit, list):
        circuit = [circuit]
    g = cnots(qubits=qubits, depth=n_cnots)
    c = Circuit.from_graph(g)
    matrix = Mat2.id(qubits)
    for gate in c.gates:
        if not hasattr(gate, "control") or not hasattr(gate, "target"):
            continue
        matrix.row_add(gate.control, gate.target)
        for c in circuit:
            c.row_add(gate.control, gate.target)
    return matrix.data


def qft(qubits: int) -> Circuit:
    """Returns a quantum Fourier transform circuit

    This returns the unoptimised Fourier transform circuit. This is the version with O(n^2) gates
    which reverses the order of logical qubits."""
    c = Circuit(qubits)

    for i in range(qubits):
        c.add_gate('H', i)
        for j in range(i+1, qubits):
            c.add_gate('CPhase', j, i, Fraction(1, 2**(j-i+1)))
    return c



# PyZX - Python library for quantum circuit rewriting 
#        and optimization using the ZX-calculus
# Copyright (C) 2018 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import sys
import numpy as np
from fractions import Fraction
from pyzx_param.graph.scalar import DyadicNumber, Scalar

if __name__ == '__main__':
    sys.path.append('..')
    sys.path.append('.')


class TestScalar(unittest.TestCase):

    def test_initialization(self):
        scalar = Scalar()
        self.assertEqual(scalar.power2, 0)
        self.assertEqual(scalar.phase, Fraction(0))
        self.assertEqual(scalar.phasenodes, [])
        self.assertEqual(scalar.floatfactor.to_complex(), 1.0)
        self.assertFalse(scalar.is_unknown)
        self.assertFalse(scalar.is_zero)

    def test_copy(self):
        scalar = Scalar()
        scalar.power2 = 2
        scalar.phase = Fraction(1, 3)
        scalar.phasenodes = [Fraction(1, 2)]
        scalar.floatfactor = DyadicNumber(0, 2, 0, 0, 0)  # 2.0
        copied_scalar = scalar.copy()
        self.assertEqual(copied_scalar.power2, 2)
        self.assertEqual(copied_scalar.phase, Fraction(1, 3))
        self.assertEqual(copied_scalar.phasenodes, [Fraction(1, 2)])
        self.assertEqual(copied_scalar.floatfactor.to_complex(), 2.0)
        self.assertFalse(copied_scalar.is_unknown)
        self.assertFalse(copied_scalar.is_zero)

    def test_conjugate(self):
        scalar = Scalar()
        scalar.phase = Fraction(1, 3)
        scalar.phasenodes = [Fraction(1, 2), Fraction(3, 4)]
        scalar.floatfactor = DyadicNumber(0, 2, 0, 1, 0)  # 2 + 1j

        conjugated_scalar = scalar.conjugate()
        self.assertEqual(conjugated_scalar.phase, -Fraction(1, 3))
        self.assertEqual(conjugated_scalar.phasenodes, [-Fraction(1, 2), -Fraction(3, 4)])
        self.assertEqual(conjugated_scalar.floatfactor.to_complex(), 2.0 - 1.0j)
        self.assertAlmostEqual(conjugated_scalar.to_number(), scalar.to_number().conjugate())

    def test_to_number(self):
        scalar = Scalar()
        scalar.phase = Fraction(1, 4)
        scalar.phasenodes = [Fraction(1, 2)]
        scalar.floatfactor = DyadicNumber(0, 2, 0, 0, 0)
        scalar.power2 = 2
        number = scalar.to_number()
        expected_number = (np.exp(1j * np.pi * 0.25) * (1 + np.exp(1j * np.pi * 0.5)) * (2 ** 0.5) ** 2) * 2.0
        self.assertAlmostEqual(number.real, expected_number.real, places=5)
        self.assertAlmostEqual(number.imag, expected_number.imag, places=5)


if __name__ == '__main__':
    unittest.main()



# PyZX - Python library for quantum circuit rewriting
#        and optimisation using the ZX-calculus
# Copyright (C) 2021 - Aleks Kissinger and John van de Wetering

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#    http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
This module contains an implementation of a genetic algorithm. As with simulated annealing, the default goal of this approach is to reduce the 2-qubit count of a fully-simplified ZX-diagram (i.e., of that circuit obtained via extraction) (see demos/LocalSearch).

Mutants are represented as pairs of (ZX-diagram, circuit) to allow for mutation operators that act on either, though by default only the two two congruences defined in congruences.py (that act at the graph-level) are used. Tournament selection is used by default but other options are available. The default fitness function is a weighted gate count defined in scores.py.
"""

from tqdm import tqdm
import numpy as np
from random import shuffle
import random
from copy import deepcopy

from .scores import wgc
from .congruences import apply_rand_lc, apply_rand_pivot
import sys
if __name__ == '__main__':
    sys.path.append('..')
from pyzx_param.simplify import to_graph_like, full_reduce
from pyzx_param.extract import extract_circuit
from pyzx_param.optimize import basic_optimization


__all__ = ['GeneticOptimizer']


class Mutant:
    """A single mutant for evolving a population via a genetic algorithm. The state of
    a mutant is a (ZX-diagram, circuit) pair where each corresponds to the other. Used by :class: `GeneticOptimizer` to evolve population.
    
    Attributes:
        c_orig: The original circuit.
        c_curr: The current circuit after mutations.
        g_curr: The current graph representation.
        score: The fitness score of the mutant.
        dead: A flag indicating if no more actions can be applied.
    """

    def __init__(self, c, g):
        self.c_orig = c
        self.c_curr = c
        self.g_curr = g.copy()
        self.score = None
        self.dead = False # no more actions can be applied to it

def default_score(m):
    """
    Calculates the default fitness score for a mutant.

    :param m: A Mutant object.
    :return: The weighted gate count of the current circuit.
    """
    return wgc(m.c_curr)


# Action accept a circuit and corresponding graph, and return (success, (new_c, new_g)). This allows for actions that act on either of graphs and circuits.
def rand_pivot(c, g, reduce_prob=0.1):
    """
    Applies a random pivot to a mutant. One of the actions used to mutate population.

    :param c: The current circuit.
    :param g: The current graph.
    :param reduce_prob: Probability of applying full reduction.
    :return: A tuple (success, (new_c, new_g)) indicating the success of the operation
             and the new circuit and graph.
    """

    g_tmp = g.copy()
    apply_rand_pivot(g_tmp)

    g_fr = g_tmp.copy()
    full_reduce(g_fr)
    c_new = extract_circuit(g_fr.copy()).to_basic_gates()
    c_new = basic_optimization(c_new)
    if random.uniform(0, 1) < reduce_prob:
        g_tmp = g_fr.copy()
    return True, (c_new, g_tmp)

def rand_lc(c, g, reduce_prob=0.1):    
    """
    Applies a random local complementation to a mutant. One of the actions used to mutate candidates.

    :param c: The current circuit.
    :param g: The current graph.
    :param reduce_prob: Probability of applying full reduction.
    :return: A tuple (success, (new_c, new_g)) indicating the success of the operation
             and the new circuit and graph.
    """

    g_tmp = g.copy()
    apply_rand_lc(g_tmp)

    g_fr = g_tmp.copy()
    full_reduce(g_fr)
    c_new = extract_circuit(g_fr.copy()).to_basic_gates()
    c_new = basic_optimization(c_new)
    if random.uniform(0, 1) < reduce_prob:
        g_tmp = g_fr.copy()
    return True, (c_new, g_tmp)


class GeneticOptimizer():
    """Implements a genetic algorithm for evolving a population of (ZX-diagram, circuit) pairs.
    The primary functionality of this class is defined in the evolve method.

    Attributes:
        actions: A list of mutation functions.
        score: A function to calculate the fitness score."""

    def __init__(self, actions=[rand_pivot, rand_lc], score=default_score):
        self.actions = actions
        self.score = score # function that maps Mutant -> Double

    # TODO: multi-thread
    def mutate(self):
        """Applies a single mutation to each member of the current population"""

        for m in self.mutants:
            # Note: actions have to look for their own matches/subjects
            success = False
            shuffle(self.actions)
            for a in self.actions:
                success, (c_new, g_new) = a(m.c_curr, m.g_curr)
                if success:
                    m.c_curr = c_new.to_basic_gates()
                    m.g_curr = g_new.copy() # copy() to make vertices consecutive
                    m.score = self.score(m)
                    break
            if not success:
                m.dead = True

    def update_scores(self):
        """Updates the fitness scores of all mutants."""
        for m in self.mutants:
            m.score = self.score(m)

    def select(self, method="tournament"):
        """Selects the next generation of mutants based on the specified method

        :param method: The selection method to use. Options are:
            - "tournament": Randomly selects pairs of mutants and chooses the one with the better score.
            - "top_half": Selects the top half of mutants based on their scores and duplicates them.
            - "top_n": Selects the top N mutants and replicates them to form the next generation.

        :raises RuntimeError: If an unknown selection method is specified.
        """

        if method == "tournament":
            new_mutants = list()
            for _ in range(self.n_mutants):
                m1, m2 = random.sample(self.mutants, 2)
                if m1.dead:
                    new_mutants.append(m2)
                elif m1.score < m2.score: # Reminder: lower is better
                    new_mutants.append(m1)
                else:
                    new_mutants.append(m2)
            self.mutants = [deepcopy(m) for m in new_mutants]
        elif method == "top_half":
            ms_tmp = self.mutants.copy()
            ms_tmp = sorted(ms_tmp, key=lambda m: m.score)
            n = self.n_mutants // 2
            top_half = ms_tmp[:n]
            if self.n_mutants % 2 == 0:
                self.mutants = deepcopy(top_half) * 2
            else:
                self.mutants = deepcopy(top_half) * 2 + deepcopy([ms_tmp[0]])
        elif method == "top_n":
            n = 10
            ms_tmp = self.mutants.copy()
            ms_tmp = sorted(ms_tmp, key=lambda m: m.score)
            top_n = ms_tmp[:n]
            fact = self.n_mutants // n
            r = self.n_mutants - fact * n
            self.mutants = top_n * fact + top_n[:r]
            # Hack to not mess up memory addresses
            self.mutants = [deepcopy(m) for m in self.mutants]
        else:
            raise RuntimeError(f"[select] Unknown selection method {method}")


    def evolve(self, g, n_mutants, n_generations, quiet=True):
        """The primary functionality of GeneticOptimizer. Given an input ZX-diagram,
        searches the space of ZX-diagrams generated by the provided action set to minimize
        the supplied fitness function.

        :param g: The initial graph.
        :param n_mutants: Number of mutants in the population.
        :param n_generations: Number of generations to evolve.
        :param quiet: If True, suppresses progress output.
        :return: The optimized graph."""

        self.quiet = quiet
        _, _, g_opt = self._evolve(g, n_mutants, n_generations)
        return g_opt

    def _evolve(self, g, n_mutants, n_generations):
        """A helper function to _evolve that tracks fitness scores over the evolution"""

        self.n_mutants = n_mutants
        self.n_gens = n_generations

        self.g_orig = g.copy()
        to_graph_like(self.g_orig)
        self.c_orig = extract_circuit(self.g_orig.copy()).to_basic_gates()
        self.c_orig = basic_optimization(self.c_orig)

        self.mutants = [Mutant(self.c_orig, self.g_orig) for _ in range(self.n_mutants)]

        self.update_scores()
        best_mutant = min(self.mutants, key=lambda m: m.score)
        best_score = best_mutant.score

        gen_scores = [best_score]
        best_scores = [best_score]
        for i in tqdm(range(self.n_gens), desc="Generations", disable=self.quiet):
            n_unique_mutants = len(list(set([id(m) for m in self.mutants])))
            assert(n_unique_mutants == self.n_mutants)

            self.mutate()
            best_in_gen = min(self.mutants, key=lambda m: m.score)
            gen_scores.append(best_in_gen.score)
            if best_in_gen.score < best_score:
                best_mutant = deepcopy(best_in_gen)
                best_score = best_in_gen.score

            best_scores.append(best_score)
            if all([m.dead for m in self.mutants]):
                break

            self.select()

        return best_scores, gen_scores, best_mutant.g_curr



# type: ignore

# NOTE: this is experimental code, and not fully implemented. Do not expect it to extract circuits.

from .utils import EdgeType, VertexType, toggle_edge
from .linalg import Mat2, Z2
from .simplify import id_simp, tcount
from .rewrite_rules.rules import apply_rule, pivot, match_spider_parallel, spider
from .circuit import Circuit
from .circuit.gates import Gate, ParityPhase, CNOT, HAD, ZPhase, XPhase, CZ, XCX, SWAP, InitAncilla
from .graph.base import BaseGraph, VT, ET, FractionLike

from typing import List, Optional, Tuple, Dict, Set, Union, Iterator

def is_clifford(g: BaseGraph[VT, ET], v: VT):
    return g.phase(v).denominator <= 2

def is_gadget(g: BaseGraph[VT, ET], v: VT):
    for n in g.neighbors(v):
        if g.type(n) == VertexType.Z and g.vertex_degree(n) == 1:
            return True
    return False

def lc_boundary(g: BaseGraph[VT, ET], q: int, c: Circuit):
    o = g.outputs()[q]
    v = next(g.neighbors(o)) # frontier and qubit q
    nhd = [w for w in g.neighbors(v) if w != o]
    for i in range(len(nhd)):
        g.add_to_phase(nhd[i], Fraction(-1,2))
        for j in range(i+1, len(nhd)):
            g.add_edge((nhd[i],nhd[j]), EdgeType.HADAMARD)
    c.prepend_gate(XPhase(Fraction(1,2)))


def gadgetize(g: BaseGraph[VT, ET]):
    verts = list(g.vertices())
    for v in verts:
        if g.type(v) == VertexType.Z and g.vertex_degree(v) > 1 and not is_clifford(g, v):
            r = g.row(v)
            q = g.qubit(v)
            v1 = g.add_vertex(VertexType.Z, qubit = -1, row = r+1)
            v2 = g.add_vertex(VertexType.Z, qubit = -2, row = r+1)
            g.add_edge((v, v1), EdgeType.HADAMARD)
            g.add_edge((v1, v2), EdgeType.HADAMARD)
            g.set_phase(v2, g.phase(v))
            g.set_phase(v, 0)


def remove_h(g: BaseGraph[VT, ET], c: Circuit):
    for q, o in enumerate(g.outputs()):
        e = next(g.incident_edges(o))
        if g.edge_type(e) == EdgeType.HADAMARD:
            c.prepend_gate(HAD(q))
            g.set_edge_type(e, EdgeType.SIMPLE)


def remove_phases(g: BaseGraph[VT, ET], c: Circuit):
    for q, o in enumerate(g.outputs()):
        v = next(g.neighbors(o)) # frontier spider
        if g.type(v) != VertexType.Z: continue
        p = g.phase(v)
        if p != 0:
            c.prepend_gate(ZPhase(p))
            g.set_phase(v, 0)

def remove_czs(g: BaseGraph[VT, ET], c: Circuit):
    outs = g.outputs()
    for i in range(len(outs)):
        v = next(g.neighbors(outs[i]))
        if g.type(v) != VertexType.Z: continue
        for j in range(i+1, len(outs)):
            w = next(g.neighbors(outs[j]))
            if g.type(w) != VertexType.Z: continue
            if g.connected(v, w):
                g.remove_edge(g.edge(v,w))
                c.prepend_gate(CZ(i, j))

def solve_gadget(g: BaseGraph[VT, ET], gd: VT):
    frontier = set(next(g.neighbors(o)) for o in g.ouputs())
    s = set(n for n in g.neighbors(gd) if g.degree(n) != 1 and not n in frontier)
    for f in frontier: s += g.neighbors(f)
    fr = list(frontier)
    nhd = list(s)
    bi_adj = Mat2([[1 if g.connected(fr[i], nhd[j]) else 0
        for i in range(len(fr))]
            for j in range(len(nhd))])
    gd_vec = Mat2([[1 if g.connected(gd, nhd[j]) else 0] for i in range(len(nhd))])
    sln = bi_adj.solve(gd_vec)
    if sln != None:
        return [fr[i] for i in range(len(fr)) if sln[i] == 1]
    else:
        return None

def clean_frontier(g: BaseGraph[VT, ET], c: Circuit):
    found_cliff = True
    while found_cliff:
        found_cliff = False
        remove_h(g, c)
        remove_phases(g, c)
        remove_czs(g, c)
        for o in g.outputs():
            v = next(g.neighbors(0))
            if g.type(v) != VertexType.Z: continue
            for n in g.neighbors(v):
                if g.type(n) == VertexType.Z and is_clifford(g, n) and not is_gadget(g, n):
                    found_cliff = True
                    break
            if found_cliff: break # break all the way out to while loop


def gadget_extract(g: BaseGraph[VT, ET]) -> Circuit:
    c = Circuit(g.num_outputs())
    gadgetize(g)
    clean_frontier(g, c)
    return c
